
from qtpy.QtCore import Qt
from qtpy.QtCore import QObject
#from qtpy.QtWidgets import QWidget
from qtpy.QtCore import QEvent
from qtpy.QtCore import QVariant
from qtpy.QtCore import Slot
from qtpy.QtCore import QItemSelection
from qtpy.QtCore import QItemSelectionModel
from qtpy.QtCore import QModelIndex
from qtpy.QtGui import QStandardItem
from qtpy.QtGui import QStandardItemModel
from qtpy.QtWidgets import QMenu


class NetworkTreeManager(QObject):

    def __init__(self, mw):
        super().__init__(parent=mw)
        self.mw = mw
        self.tree_view = self.mw.ui_navigationTree
        model = self.mw.nm_manager.model

        self.tree_view.setRootIsDecorated(True)
        self.tree_view.setHeaderHidden(True)
        if model:
            self.tree_view.setModel(model)
            self.tree_view.expandAll()

            selection_model = self.tree_view.selectionModel()
            selection_model.currentChanged.connect(self.on_current_changed)
            selection_model.selectionChanged.connect(self.on_selection_changed)

        self.tree_view.keyPressed.connect(self.keyPressEvent)

        # Create context menu
        self.tree_view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tree_view.customContextMenuRequested.connect(self.on_context_menu)
        self.pop_menu = QMenu(self.tree_view)

        self.action_open_network = self.pop_menu.addAction('Open Network')
        self.action_save_network = self.pop_menu.addAction('Save Network')
        self.action_remove_selected = self.pop_menu.addAction('Remove Selected')
        self.action_remove_selected.setEnabled(False)

    @property
    def tree_view(self):
        return self._tree_view

    @tree_view.setter
    def tree_view(self, tv):
        self._tree_view = tv

    @property
    def current_item(self):
        # indexes = self.history_view.selectedIndexes()
        # if len(indexes) >= 1:
        #     idx = indexes[-1]
        #     model = self.history_view.model()
        #     item = model.itemFromIndex(idx)
        #     return item

        ind = self.tree_view.currentIndex()
        if ind.row() >= 0:
            model = self.tree_view.model()
            item = model.itemFromIndex(ind)
            return item

        return None

    @property
    def current_net(self):
        item = self.current_item
        if not item:
            return None
        net = item.data()
        return net

    @Slot(QEvent)
    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Delete:
            self.remove_selected_items()

    @Slot(QModelIndex, QModelIndex)
    def on_current_changed(self, current, previous):
        if current.row() < 0:
            self.mw.sv_manager.clear()
            self.mw.ct_manager.update_console_vars()
            return

        model = self.tree_view.model()
        item = model.itemFromIndex(current)
        net = item.data()
        scene = net.scene
        self.action_remove_selected.setEnabled(True)
        self.mw.sv_manager.set_current_view_scene(net.scene, net.name)
        self.mw.ct_manager.update_console_vars()

        self.mw.hv_manager.update_history_view(scene.history)

        # # Process previous net
        # if previous.row() >= 0:
        #     item = model.itemFromIndex(previous)
        #     net = item.data()
        #     scene = net.scene
        #     scene.history.hide_history_view()

    @Slot(QItemSelection, QItemSelection)
    def on_selection_changed(self, selected, deselected):
        self.mw.sv_manager.view.enable_menu_align()

    def append_item(self, net):
        item = QStandardItem(net.name)
        item.setData(QVariant(net))
        model = self.tree_view.model()
        root = model.invisibleRootItem()
        root.appendRow(item)
        net.item = item
        self.tree_view.setCurrentIndex(item.index())
        """
        Following is a code snippet to consult.

        selection_model = self.history_view.selectionModel()
        selection_model.select(item.index(),
                              QItemSelectionModel.ClearAndSelect)
        """

    def clear_selection(self):
        selection_model = self.tree_view.selectionModel()
        selection_model.clear_selection()

    def on_context_menu(self, point):
        action_selected = self.pop_menu.exec_(self.tree_view.mapToGlobal(point))
        if action_selected == self.action_open_network:
            self.mw.ma_handler.process_open_network()
        elif action_selected == self.action_save_network:
            self.mw.ma_handler.process_save_network()
        elif action_selected == self.action_remove_selected:
            self.remove_selected_items()

    def remove_selected_items(self):
        model = self.tree_view.model()
        root = model.invisibleRootItem()
        indexes = self.tree_view.selectedIndexes()

        while len(indexes) > 0:
            ind = indexes[0]
            model.removeRow(ind.row(), ind.parent())
            indexes = self.tree_view.selectedIndexes()

        if root.rowCount() == 0:
            self.mw.sv_manager.clear()
            self.mw.hv_manager.clear()
        else:
            self.on_selection_changed([], [])


    def clear(self):
        self.tree_view.selectAll()
        self.remove_selected_items()


import math
from qtpy.QtGui import QPainterPath

from nezzle.graphics.mixins import Lockable
from nezzle.graphics.nodes.basenode import BaseNode


@Lockable
class RectangleNode(BaseNode):

    ITEM_TYPE = 'RECTANGLE_NODE'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # end of def __init__

    def shape(self):
        path = QPainterPath()
        path.addRect(self._brect)
        return path

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)
        painter.drawRect(self._brect)

    def calculate_radius(self, angle):
        """
        The angle should be in radians.
        """
        a = self.width/2
        b = self.height/2
        return 1/math.sqrt(1/b**2 + (1/a**2 - 1/b**2)*(math.cos(angle)**2))


@Lockable
class SquareNode(RectangleNode):

    ITEM_TYPE = 'SQUARE_NODE'

    def __init__(self, iden, radius, *args, **kwargs):
        super().__init__(iden, width=2*radius, height=2*radius, *args, **kwargs)
        self._attr.set_trigger('RADIUS', self._trigger_set_radius)
        self._attr['RADIUS'] = radius
        assert self.width == self.height
    # end of def __init__

    def _trigger_set_radius(self, key, value):
        diameter = 2*value
        self._attr['WIDTH'] = diameter
        self._attr['HEIGHT'] = diameter
        return value

    @property
    def width(self):
        return self._attr['WIDTH']

    @width.setter
    def width(self, val):
        self._attr['RADIUS'] = val/2

    @property
    def height(self):
        return self._attr['HEIGHT']

    @height.setter
    def height(self, val):
        self._attr['RADIUS'] = val/2

    @property
    def radius(self):
        return self._attr['RADIUS']

    @radius.setter
    def radius(self, val):
        self._attr['RADIUS'] = val

    def calculate_radius(self, angle=None):
        return self.radius

    @classmethod
    def from_dict(cls, attr):
        iden = attr.pop('ID')
        radius = attr.pop('RADIUS')

        obj = cls(iden=iden, radius=radius)
        obj._attr.update(attr)

        return obj



import nezzle


def search_links(net, substr):
    """Find links whose identity string contain the given substring.

    Args:
        net: nezzle.graphics.Network
            Network object whose links are searched.
        substr: str
            Substring to search in the identity string of a link (link.iden).

    Returns:
        links: list
            List of links whose identity strings contain the substring.

    """
    links_found = []
    for iden, link in net.links.items():
        if isinstance(link, nezzle.graphics.SelfloopLink):
            if substr in link.node.iden:
                links_found.append(link)
            else:
                continue
        elif substr in link.target.iden or substr in link.source.iden:
            links_found.append(link)

    return links_found


def search_links_target(net, substr):
    links_found = []
    for src, tgt, data in net.nxdg.edges(data=True):
        link = data['GRAPHICS']
        if isinstance(link, nezzle.graphics.SelfloopLink):
            if substr in link.node:
                links_found.append(link)
        elif substr in link.target.iden:
            links_found.append(link)
    # end of for
    return links_found


from os.path import dirname
from os.path import join as pjoin

__author__ = "Daewon Lee"
__copyright__ = "Copyright 2016-2021, Daewon Lee, All Rights Reserved."
__credits__ = ["Daewon Lee",]
__license__ = "MIT"
__maintainer__ = "Daewon Lee (daewon4you@gmail.com)"
__email__ = "daewon4you@gmail.com"
__status__ = "Prototype"

with open(pjoin(dirname(__file__), "VERSION"), "rt") as fin:
    __version__ = fin.read().strip()


from nezzle.fileio import read_sif
from nezzle.fileio import write_sif

from nezzle.fileio import read_json
from nezzle.fileio import write_json

from nezzle.fileio import read_network
from nezzle.fileio import write_network

from nezzle.convert import to_graphics
from nezzle.convert import to_networkx


from qtpy.QtWidgets import QGraphicsItem

from nezzle.systemstate import get_system_state


def Lockable(__class):

    class __Lockable(__class):

        def mousePressEvent(self, event):
            ss = get_system_state()
            if ss.is_locked(self.ITEM_TYPE):
                # event.ignore()
                event.ignore()
            else:
                return super().mousePressEvent(event)

        def itemChange(self, change, value):
            if change == QGraphicsItem.ItemSelectedChange:
                ss = get_system_state()
                if ss.is_locked(self.ITEM_TYPE):
                    return super().itemChange(change, False)

            return super().itemChange(change, value)

        def __repr__(self):
            return '<Lockable %s object at %s>'\
                   %(self.__class__.__name__, hex(id(self)))

        # def __str__(self):
        #     return super().__str__()

    __Lockable.__name__ = __class.__name__
    return __Lockable



from qtpy.QtCore import Qt
from qtpy.QtCore import Signal
from qtpy.QtCore import QEvent
from qtpy.QtWidgets import QTreeView


class NetworkTreeView(QTreeView):
    keyPressed = Signal(QEvent)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        self.keyPressed.emit(event)


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './nezzle/ui/ui_exportimagedialog.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ExportImageDialog(object):
    def setupUi(self, ExportImageDialog):
        ExportImageDialog.setObjectName("ExportImageDialog")
        ExportImageDialog.resize(344, 406)
        ExportImageDialog.setSizeGripEnabled(False)
        ExportImageDialog.setModal(True)
        self.layoutWidget = QtWidgets.QWidget(ExportImageDialog)
        self.layoutWidget.setGeometry(QtCore.QRect(0, 0, 2, 2))
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.layoutWidget1 = QtWidgets.QWidget(ExportImageDialog)
        self.layoutWidget1.setGeometry(QtCore.QRect(0, 0, 2, 2))
        self.layoutWidget1.setObjectName("layoutWidget1")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.layoutWidget2 = QtWidgets.QWidget(ExportImageDialog)
        self.layoutWidget2.setGeometry(QtCore.QRect(0, 0, 2, 2))
        self.layoutWidget2.setObjectName("layoutWidget2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.layoutWidget2)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.layoutWidget3 = QtWidgets.QWidget(ExportImageDialog)
        self.layoutWidget3.setGeometry(QtCore.QRect(0, 0, 2, 2))
        self.layoutWidget3.setObjectName("layoutWidget3")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.layoutWidget3)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.layoutWidget4 = QtWidgets.QWidget(ExportImageDialog)
        self.layoutWidget4.setGeometry(QtCore.QRect(12, 8, 321, 388))
        self.layoutWidget4.setObjectName("layoutWidget4")
        self.verticalLayout_13 = QtWidgets.QVBoxLayout(self.layoutWidget4)
        self.verticalLayout_13.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_13.setObjectName("verticalLayout_13")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.label = QtWidgets.QLabel(self.layoutWidget4)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.ui_filePathEdit = QtWidgets.QLineEdit(self.layoutWidget4)
        self.ui_filePathEdit.setReadOnly(False)
        self.ui_filePathEdit.setObjectName("ui_filePathEdit")
        self.horizontalLayout.addWidget(self.ui_filePathEdit)
        self.ui_openButton = QtWidgets.QPushButton(self.layoutWidget4)
        self.ui_openButton.setObjectName("ui_openButton")
        self.horizontalLayout.addWidget(self.ui_openButton)
        spacerItem1 = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout_13.addLayout(self.horizontalLayout)
        self.groupBox_5 = QtWidgets.QGroupBox(self.layoutWidget4)
        self.groupBox_5.setFlat(False)
        self.groupBox_5.setCheckable(False)
        self.groupBox_5.setObjectName("groupBox_5")
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout(self.groupBox_5)
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.ui_transparencyCheck = QtWidgets.QCheckBox(self.groupBox_5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_transparencyCheck.sizePolicy().hasHeightForWidth())
        self.ui_transparencyCheck.setSizePolicy(sizePolicy)
        self.ui_transparencyCheck.setObjectName("ui_transparencyCheck")
        self.horizontalLayout_5.addWidget(self.ui_transparencyCheck)
        spacerItem2 = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem2)
        self.horizontalLayout_13.addLayout(self.horizontalLayout_5)
        self.verticalLayout_13.addWidget(self.groupBox_5)
        spacerItem3 = QtWidgets.QSpacerItem(20, 2, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.verticalLayout_13.addItem(spacerItem3)
        self.groupBox_4 = QtWidgets.QGroupBox(self.layoutWidget4)
        self.groupBox_4.setFlat(False)
        self.groupBox_4.setCheckable(False)
        self.groupBox_4.setObjectName("groupBox_4")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.groupBox_4)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.ui_qualitySlider = QtWidgets.QSlider(self.groupBox_4)
        self.ui_qualitySlider.setOrientation(QtCore.Qt.Horizontal)
        self.ui_qualitySlider.setObjectName("ui_qualitySlider")
        self.horizontalLayout_12.addWidget(self.ui_qualitySlider)
        self.ui_qualityEdit = QtWidgets.QLineEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_qualityEdit.sizePolicy().hasHeightForWidth())
        self.ui_qualityEdit.setSizePolicy(sizePolicy)
        self.ui_qualityEdit.setMaximumSize(QtCore.QSize(30, 16777215))
        self.ui_qualityEdit.setText("")
        self.ui_qualityEdit.setDragEnabled(True)
        self.ui_qualityEdit.setObjectName("ui_qualityEdit")
        self.horizontalLayout_12.addWidget(self.ui_qualityEdit)
        self.ui_qualityPercentLabel = QtWidgets.QLabel(self.groupBox_4)
        self.ui_qualityPercentLabel.setObjectName("ui_qualityPercentLabel")
        self.horizontalLayout_12.addWidget(self.ui_qualityPercentLabel)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_12)
        self.verticalLayout_13.addWidget(self.groupBox_4)
        spacerItem4 = QtWidgets.QSpacerItem(20, 3, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.verticalLayout_13.addItem(spacerItem4)
        self.groupBox_3 = QtWidgets.QGroupBox(self.layoutWidget4)
        self.groupBox_3.setFlat(False)
        self.groupBox_3.setCheckable(False)
        self.groupBox_3.setObjectName("groupBox_3")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.groupBox_3)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.ui_scaleWidthLabel = QtWidgets.QLabel(self.groupBox_3)
        self.ui_scaleWidthLabel.setObjectName("ui_scaleWidthLabel")
        self.horizontalLayout_10.addWidget(self.ui_scaleWidthLabel)
        self.ui_scaleWidthEdit = QtWidgets.QLineEdit(self.groupBox_3)
        self.ui_scaleWidthEdit.setDragEnabled(True)
        self.ui_scaleWidthEdit.setObjectName("ui_scaleWidthEdit")
        self.horizontalLayout_10.addWidget(self.ui_scaleWidthEdit)
        self.horizontalLayout_9.addLayout(self.horizontalLayout_10)
        self.ui_scaleWidthPercentLabel = QtWidgets.QLabel(self.groupBox_3)
        self.ui_scaleWidthPercentLabel.setObjectName("ui_scaleWidthPercentLabel")
        self.horizontalLayout_9.addWidget(self.ui_scaleWidthPercentLabel)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        spacerItem5 = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_11.addItem(spacerItem5)
        self.ui_scaleHeightLabel = QtWidgets.QLabel(self.groupBox_3)
        self.ui_scaleHeightLabel.setObjectName("ui_scaleHeightLabel")
        self.horizontalLayout_11.addWidget(self.ui_scaleHeightLabel)
        self.ui_scaleHeightEdit = QtWidgets.QLineEdit(self.groupBox_3)
        self.ui_scaleHeightEdit.setDragEnabled(True)
        self.ui_scaleHeightEdit.setObjectName("ui_scaleHeightEdit")
        self.horizontalLayout_11.addWidget(self.ui_scaleHeightEdit)
        self.horizontalLayout_9.addLayout(self.horizontalLayout_11)
        self.ui_scaleHeightPercentLabel = QtWidgets.QLabel(self.groupBox_3)
        self.ui_scaleHeightPercentLabel.setObjectName("ui_scaleHeightPercentLabel")
        self.horizontalLayout_9.addWidget(self.ui_scaleHeightPercentLabel)
        self.horizontalLayout_3.addLayout(self.horizontalLayout_9)
        self.verticalLayout_13.addWidget(self.groupBox_3)
        spacerItem6 = QtWidgets.QSpacerItem(20, 3, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.verticalLayout_13.addItem(spacerItem6)
        self.groupBox_2 = QtWidgets.QGroupBox(self.layoutWidget4)
        self.groupBox_2.setFlat(False)
        self.groupBox_2.setCheckable(False)
        self.groupBox_2.setObjectName("groupBox_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.groupBox_2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_6 = QtWidgets.QLabel(self.groupBox_2)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_6.addWidget(self.label_6)
        self.ui_dpiWidthEdit = QtWidgets.QLineEdit(self.groupBox_2)
        self.ui_dpiWidthEdit.setDragEnabled(True)
        self.ui_dpiWidthEdit.setObjectName("ui_dpiWidthEdit")
        self.horizontalLayout_6.addWidget(self.ui_dpiWidthEdit)
        self.ui_dpiWidthLabel = QtWidgets.QLabel(self.groupBox_2)
        self.ui_dpiWidthLabel.setObjectName("ui_dpiWidthLabel")
        self.horizontalLayout_6.addWidget(self.ui_dpiWidthLabel)
        self.horizontalLayout_8.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        spacerItem7 = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem7)
        self.label_7 = QtWidgets.QLabel(self.groupBox_2)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_7.addWidget(self.label_7)
        self.ui_dpiHeightEdit = QtWidgets.QLineEdit(self.groupBox_2)
        self.ui_dpiHeightEdit.setDragEnabled(True)
        self.ui_dpiHeightEdit.setObjectName("ui_dpiHeightEdit")
        self.horizontalLayout_7.addWidget(self.ui_dpiHeightEdit)
        self.ui_dpiHeightLabel = QtWidgets.QLabel(self.groupBox_2)
        self.ui_dpiHeightLabel.setObjectName("ui_dpiHeightLabel")
        self.horizontalLayout_7.addWidget(self.ui_dpiHeightLabel)
        self.horizontalLayout_8.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_2.addLayout(self.horizontalLayout_8)
        self.verticalLayout_13.addWidget(self.groupBox_2)
        spacerItem8 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.verticalLayout_13.addItem(spacerItem8)
        self.ui_buttonBox = QtWidgets.QDialogButtonBox(self.layoutWidget4)
        self.ui_buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.ui_buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.ui_buttonBox.setObjectName("ui_buttonBox")
        self.verticalLayout_13.addWidget(self.ui_buttonBox)

        self.retranslateUi(ExportImageDialog)
        self.ui_buttonBox.accepted.connect(ExportImageDialog.accept)
        self.ui_buttonBox.rejected.connect(ExportImageDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(ExportImageDialog)

    def retranslateUi(self, ExportImageDialog):
        _translate = QtCore.QCoreApplication.translate
        ExportImageDialog.setWindowTitle(_translate("ExportImageDialog", "Dialog"))
        self.label.setText(_translate("ExportImageDialog", "File"))
        self.ui_openButton.setText(_translate("ExportImageDialog", "Open"))
        self.groupBox_5.setTitle(_translate("ExportImageDialog", "Background"))
        self.ui_transparencyCheck.setText(_translate("ExportImageDialog", "Transparent Background"))
        self.groupBox_4.setToolTip(_translate("ExportImageDialog", "100% means the image is not compressed."))
        self.groupBox_4.setTitle(_translate("ExportImageDialog", "Quality"))
        self.ui_qualityPercentLabel.setText(_translate("ExportImageDialog", "%"))
        self.groupBox_3.setTitle(_translate("ExportImageDialog", "Scale"))
        self.ui_scaleWidthLabel.setText(_translate("ExportImageDialog", "Width:"))
        self.ui_scaleWidthPercentLabel.setText(_translate("ExportImageDialog", "%"))
        self.ui_scaleHeightLabel.setText(_translate("ExportImageDialog", "Height:"))
        self.ui_scaleHeightPercentLabel.setText(_translate("ExportImageDialog", "%"))
        self.groupBox_2.setTitle(_translate("ExportImageDialog", "DPI"))
        self.label_6.setText(_translate("ExportImageDialog", "Width:"))
        self.ui_dpiWidthLabel.setText(_translate("ExportImageDialog", "dpi"))
        self.label_7.setText(_translate("ExportImageDialog", "Height:"))
        self.ui_dpiHeightLabel.setText(_translate("ExportImageDialog", "dpi"))




from qtpy.QtCore import Signal
from qtpy.QtCore import QItemSelection
from qtpy.QtWidgets import QUndoView


class HistoryView(QUndoView):
    itemSelectionChanged = Signal(QItemSelection, QItemSelection)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle("History")



from qtpy.QtCore import QObject
from nezzle.constants import Lock


class SystemState(QObject):

    def __init__(self):

        self._lock_state = 0

    @property
    def lock_state(self):
        return self._lock_state

    def is_locked(self, item_type):

        if isinstance(item_type, str):
            item_type = item_type.casefold()
            if item_type.endswith('_node'):
                lock = Lock.NODES
            elif item_type.endswith('_link'):
                lock = Lock.LINKS
            elif item_type.endswith('_label'):
                lock = Lock.LABELS
        elif isinstance(item_type, int):
            lock = item_type
        else:
            raise TypeError("item_should be a str or int, not %s"%item_type)

        return (self.lock_state & lock) != 0

    def set_locked(self, lock, b):

        if b is True:
            self._lock_state |= lock
        else:
            self._lock_state &= (~lock)

# end of class SystemState


system_state = SystemState()

def get_system_state():
    global system_state
    return system_state


from datetime import datetime

from qtpy.QtCore import QPointF
from qtpy.QtWidgets import QUndoCommand

from nezzle.graphics.nodes.basenode import BaseNode
from nezzle.graphics.links.controlpoint import ControlPoint
from nezzle.graphics.links.controlpoint import ConnectorControlPoint
from nezzle.graphics.links.linkfactory import LinkClassFactory
from nezzle.graphics.nodes.nodefactory import NodeClassFactory


class MoveByMouseCommand(QUndoCommand):
    ID = 0

    def __init__(self, items, parent=None):
        super().__init__(parent)
        self._items = items
        self._new_positions = []
        self._old_positions = []

        for item in items:
            self._new_positions.append((item, QPointF(item.pos())))
            self._old_positions.append((item, QPointF(item["_OLD_POS"])))

            children = item.childItems()
            if not children:
                children = []

            if isinstance(item, BaseNode):
                for link in item.links:
                    for child in link.childItems():
                        if child not in children:
                            children.append(child)

            if children:
                for child in children:
                    self._new_positions.append((child, QPointF(child.pos())))
                    self._old_positions.append((child, QPointF(child["_OLD_POS"])))

        # end of for

    def id(self):
        return self.ID

    def undo(self):
        item = self._items[0]
        scene = item.scene()
        selected = scene.selectedItems()
        scene.clearSelection()
        scene.clearFocus()

        for i, (item, pos) in enumerate(self._old_positions):

            if isinstance(item, ConnectorControlPoint):
                item.setSelected(True)
                item.setPos(pos)
                item.parent.update()
                item.setSelected(False)
            else:
                item.setPos(pos)

            new_item, new_pos = self._new_positions[i]

        for selected_item in selected:
            selected_item.setSelected(True)

        scene.update()
        self.setText(self._text)

    def redo(self):
        for i, (item, pos) in enumerate(self._new_positions):
            if isinstance(item, ConnectorControlPoint):
                old_selected = item.isSelected()
                item.setSelected(True)
                item.setPos(pos)
                item.parent.update()
                item.setSelected(old_selected)
            else:
                item.setPos(pos)

        # end of for
        item.scene().update()

        num_items = len(self._items)
        if num_items == 1:
            item = self._items[0]
            _, old_pos = self._old_positions[0]
            self._text = f"Move {str(item)} from ({old_pos.x():.2f}, {old_pos.y():.2f}) to ({item.x():.2f}, {item.y():.2f})"
        else:
            if isinstance(self._items[0], ControlPoint):
                self._text = f"Move {num_items} control points"
            else:
                self._text = f"Move {num_items} Items"

        self.setText(self._text)


class MoveByKeyCommand(MoveByMouseCommand):
    ID = 1

    def id(self):
        return self.ID
    """
    The mergeWith() makes the consecutive moves a single MoveCommand,
    (i.e., the item will be moved back to the first position,
          when undo is performed.)
    """
    def mergeWith(self, command):

        if not isinstance(command, MoveByKeyCommand):
            return False

        items = command._items
        self._new_positions = [(item, QPointF(item.pos())) for item in items]

        num_items = len(self._items)
        if num_items == 1:
            item = self._items[0]
            old_pos = item.pos()
            _, new_pos = self._new_positions[0]
            self._text = f"Move {str(item)} from ({old_pos.x():.2f}, {old_pos.y():.2f}) to ({new_pos.x():.2f}, {new_pos.y():.2f})"
        else:
            self._text = f"Move {num_items} Items"
        self.setText(self._text)
        return True


class ConvertNodeCommand(QUndoCommand):
    ID = 2

    def __init__(self, net, old_items, new_items, parent=None):
        super().__init__(parent)
        self._net = net
        self._old_items = old_items
        self._new_items = new_items

    def undo(self):
        for old_item, new_item in zip(self._old_items, self._new_items):
            old_item.setSelected(new_item.isSelected())
            self._net.replace_node(new_item, old_item)
            old_item.update_children_old_positions()
            old_item["_OLD_POS"] = old_item.pos()

        old_item.scene().update()
        self.setText(self._text)

    def redo(self):
        for old_item, new_item in zip(self._old_items, self._new_items):
            new_item.setSelected(old_item.isSelected())
            self._net.replace_node(old_item, new_item)
            new_item.update_children_old_positions()
            new_item["_OLD_POS"] = new_item.pos()

        new_item.scene().update()

        NewNodeClass = NodeClassFactory.create(new_item.ITEM_TYPE)
        num_items = len(self._old_items)
        if num_items == 1:
            old_item = self._old_items[0]
            new_item = self._new_items[0]

            self._text = f"Convert {str(old_item)} to {NewNodeClass.__name__}"
        else:
            self._text = f"Convert {num_items} items to {NewNodeClass.__name__}"

        self.setText(self._text)


class ConvertLinkCommand(QUndoCommand):
    ID = 3

    def id(self):
        return self.ID

    def __init__(self, net, old_items, new_items, parent=None):
        super().__init__(parent)
        self._net = net
        self._old_items = old_items  # [item.copy() for item in new_items]
        self._new_items = new_items  # [item.copy() for item in old_items]

    def undo(self):
        for old_item, new_item in zip(self._old_items, self._new_items):
            old_item.setSelected(new_item.isSelected())
            self._net.replace_link(new_item, old_item)
            old_item.update_children_old_positions()
            old_item["_OLD_POS"] = new_item.pos()

        old_item.scene().update()
        self.setText(self._text)

    def redo(self):
        for old_item, new_item in zip(self._old_items, self._new_items):
            new_item.setSelected(old_item.isSelected())
            self._net.replace_link(old_item, new_item)
            new_item.update_children_old_positions()
            new_item["_OLD_POS"] = new_item.pos()

        new_item.scene().update()

        NewLinkClass = LinkClassFactory.create(new_item.ITEM_TYPE)
        num_items = len(self._old_items)
        if num_items == 1:
            old_item = self._old_items[0]
            new_item = self._new_items[0]
            self._text = f"Convert {str(old_item)} to {NewLinkClass.__name__}"
        else:
            self._text = f"Convert {num_items} Items to {NewLinkClass.__name__}"

        self.setText(self._text)



from nezzle.graphics.labels.textlabel import TextLabel


class LabelClassFactory(object):

    @staticmethod
    def create(label_type):
        if label_type.upper() == 'TEXT_LABEL':
            return TextLabel
        else:
            raise TypeError("Undefined labels type: %s" % (label_type))



from typing import AnyStr
from typing import Union
from typing import Type

from nezzle.graphics.links.baselink import BaseLink
from nezzle.graphics.links.baselink import TwoNodeLink
from nezzle.graphics.links.straightlink import StraightLink
from nezzle.graphics.links.curvedlink import CurvedLink
from nezzle.graphics.links.elbowlink import VerticalElbowLink, HorizontalElbowLink
from nezzle.graphics.links.linkfactory import LinkClassFactory


class LinkConverter(object):

    @staticmethod
    def convert(link: BaseLink, link_type: Union[Type, AnyStr]):

        if isinstance(link_type, str):
            link_type = LinkClassFactory.create(link_type)

        if type(link) == link_type:
            return

        attr = link.to_dict()
        attr["ITEM_TYPE"] = link_type.ITEM_TYPE
        new_link = link_type.from_dict(attr=attr, source=link.source, target=link.target)
        return new_link




from nezzle.utils.math import rotate
from nezzle.utils.math import reflect_yaxis, reflect_xaxis


class BaseTransform(object):

    def __call__(self, pos_point, pos_head=None):
        raise NotImplemented()


class Rotate(BaseTransform):
    def __init__(self, angle=None):
        self._angle = angle

    def __call__(self, pos_point, pos_head=None):
        return rotate(pos_head, pos_point, self._angle)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, val):
        self._angle = val


class Reflect(BaseTransform):
    def __init__(self, axis):
        if axis in [0, "y"]:
            self._reflect = reflect_yaxis
        elif axis in [1, "x"]:
            self._reflect = reflect_xaxis

    def __call__(self, pos_point, pos_head=None):
        return self._reflect(pos_head, pos_point)


from qtpy.QtGui import QStandardItem
from qtpy.QtGui import QStandardItemModel
from qtpy.QtCore import QObject
from qtpy.QtCore import Slot
from qtpy.QtCore import Signal


class NetworkModelManager(QObject):

    def __init__(self, mainWindow, model=None):
        super().__init__()
        self.mw = mainWindow
        if not model:
            model = QStandardItemModel()
            model.setColumnCount(1)

        self._model = model
        self._root = self.model.invisibleRootItem()
        self._model.itemChanged.connect(self.on_item_changed)

    @property
    def model(self):
        return self._model

    @property
    def root(self):
        return self._root

    @Slot(QStandardItem)
    def on_item_changed(self, item):
        net = item.data()
        if net.name != item.text():
            net.name = item.text()

        if item == self.mw.nt_manager.current_item:
            self.mw.sv_manager.set_current_view_text(net.name)



from nezzle.graphics.baseitem import MappableGraphicsItem
from nezzle.graphics.baseitem import PainterOptionItem
from nezzle.graphics.nodes.nodefactory import NodeClassFactory
from nezzle.graphics.nodes.basenode import BaseNode
from nezzle.graphics.nodes.ellipsenode import EllipseNode
from nezzle.graphics.nodes.ellipsenode import CircleNode
from nezzle.graphics.nodes.rectanglenode import RectangleNode
from nezzle.graphics.arrows.arrowclassfactory import ArrowClassFactory
from nezzle.graphics.arrows.basearrow import Triangle
from nezzle.graphics.arrows.basearrow import Hammer
from nezzle.graphics.links.linkfactory import LinkClassFactory
from nezzle.graphics.links.baselink import BaseLink
from nezzle.graphics.links.straightlink import StraightLink
from nezzle.graphics.links.curvedlink import CurvedLink
from nezzle.graphics.links.elbowlink import VerticalElbowLink
from nezzle.graphics.links.elbowlink import HorizontalElbowLink
from nezzle.graphics.links.selflooplink import SelfloopLink
from nezzle.graphics.labels.labelfactory import LabelClassFactory
from nezzle.graphics.labels.textlabel import TextLabel
from nezzle.graphics.network import Network
from nezzle.graphics.network import from_adj_to_net
from nezzle.graphics.screen import GraphicsScene
from nezzle.graphics.screen import GraphicsView



from qtpy.QtCore import Signal
from qtpy.QtWidgets import QDockWidget


class DockWidget(QDockWidget):
    closeEventOccured = Signal(bool)

    def closeEvent(self, event):
        self.closeEventOccured.emit(False)
        return super().closeEvent(event)



from qtpy.QtCore import Qt
from qtpy.QtGui import QFont
from qtpy.QtGui import QColor

from qtpy.QtWidgets import QGraphicsItem
from qtpy.QtWidgets import QGraphicsTextItem

from nezzle.graphics.mixins import Lockable
from nezzle.graphics.baseitem import PainterOptionItem
from nezzle.graphics import BaseLink


@Lockable
class TextLabel(PainterOptionItem):

    ITEM_TYPE = 'TEXT_LABEL'

    def __init__(self, parent, text, font=None, pos=None):
        self._text_item = QGraphicsTextItem()
        self._font = None

        super().__init__(parent=parent, iden=text)

        self._attr.set_trigger('TEXT',
                               self._trigger_set_text_str, when='set')
        self._attr.set_trigger('TEXT_COLOR',
                               self._trigger_set_text_color, when='set')
        self._attr.set_trigger('FONT',
                               self._trigger_set_font, when='set')
        self._attr.set_trigger('FONT_SIZE',
                               self._trigger_set_font_size, when='set')
        self._attr.set_trigger('FONT_FAMILY',
                               self._trigger_set_font_family, when='set')
        self._attr.set_trigger('FONT_BOLD',
                               self._trigger_set_font_bold, when='set')
        self._attr.set_trigger('FONT_ITALIC',
                               self._trigger_set_font_italic, when='set')

        if not font:
            font = QFont()
            font.setFamily('Arial')

        self.font = font

        if pos:
            self.setPos(pos)
            self._text_item.setPos(pos)

        self._text_item.setDefaultTextColor(Qt.black)

        self.setFlags(QGraphicsItem.ItemIsSelectable
                      | QGraphicsItem.ItemIsMovable
                      | QGraphicsItem.ItemIsFocusable
                      | QGraphicsItem.ItemSendsGeometryChanges)

        self._attr['TEXT'] = text

        self.update()

    # end of def __init__

    def __str__(self):
        return 'TextLabel(%s) belongs to %s'%(self.name, self.parentItem())

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, obj):
        self._attr['FONT'] = obj

    @property
    def text(self):
        return self._attr['TEXT']

    @text.setter
    def text(self, val):
        self._attr['TEXT'] = val

    @property
    def text(self):
        return self._attr['TEXT']

    @text.setter
    def text(self, val):
        self._attr['TEXT'] = val

    @property
    def text_item(self):
        return self._text_item

    def _trigger_set_x(self, key, value):
        self.setX(value)
        self._text_item.setX(value)
        self.update()
        return value

    def _trigger_set_y(self, key, value):
        self.setY(value)
        self._text_item.setY(value)
        self.update()
        return value

    def _trigger_set_font(self, key, value):

        if isinstance(value, QFont):
            self._font = QFont(value)
        elif isinstance(value, str):
            self._font = QFont()
            self._font.fromString(value)
        else:
            fstr = "%s is not allowed for setting font." % (type(value))
            raise ValueError(fstr)

        # Synchronize the properties related to the font.
        self._attr.set('FONT_SIZE', self._font.pixelSize(), trigger=False)
        self._attr.set('FONT_FAMILY', self._font.family(), trigger=False)
        self._attr.set('FONT_BOLD', self._font.bold(), trigger=False)
        self._attr.set('FONT_ITALIC', self._font.italic(), trigger=False)

        # update should be called to reflect the font
        # to that of self._text_item
        self.update()
        return self._font

    def _trigger_set_text_str(self, key, value):
        self._text_item.setPlainText(value)
        self.update()
        return value

    def _trigger_set_text_color(self, key, value):
        color = QColor(value)
        self._text_item.setDefaultTextColor(color)
        self.update()
        return color.name(QColor.HexArgb)

    def _trigger_set_font_size(self, key, value):
        self.font.setPixelSize(value)
        self.update()
        return value

    def _trigger_set_font_family(self, key, value):
        self.font.setFamily(value)
        self.update()
        return value

    def _trigger_set_font_bold(self, key, value):
        self.font.setBold(value)
        self.update()
        return value

    def _trigger_set_font_italic(self, key, value):
        self.font.setItalic(value)
        self.update()
        return value

    def setSelected(self, val):
        self._text_item.setSelected(val)
        super().setSelected(val)

    def boundingRect(self):
        return self._text_item.boundingRect()

    def shape(self):
        return self._text_item.shape()

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)
        self._text_item.paint(painter, option, widget)

    def is_movable(self):
        return True

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            parent = self.parentItem()
            if isinstance(parent, BaseLink):
                if parent.is_node_selected():
                    self._text_item.itemChange(change, self.pos())
                    return super().itemChange(change, self.pos())

        return super().itemChange(change, value)

    def update(self):
        if self._font:
            """ update() can be called although self._font is not initialized.
                For example, __init__ of a super class calls it
                to initialize the position.
            """
            self._text_item.setFont(self._font)
            self._text_item.update()
        self._invalidate()
        super().update()

    def copy(self, parent=None):
        if not parent:
            parent = self.parentItem()

        return self.from_dict(self.to_dict(), parent)

    def to_dict(self):
        attr = super().to_dict()
        attr['ID_PARENT'] = self.parentItem().iden
        attr['FONT'] = self._text_item.font().toString()
        return attr

    @classmethod
    def from_dict(cls, attr, parent):
        text = attr.pop('TEXT')

        strfont = attr.pop('FONT')
        font = QFont()
        font.fromString(strfont)

        obj = cls(parent, text, font)
        obj._attr.update(attr)
        return obj

    def align(self, mode="middle-center"):
        rect = self._text_item.boundingRect()

        if "top" in mode:
            self._attr["POS_Y"] = -rect.height()
        elif "bottom" in mode:
            self._attr["POS_Y"] = 0
        elif "middle" in mode:
            self._attr["POS_Y"] = -rect.height() / 2

        if "left" in mode:
            self._attr["POS_X"] = -rect.width()
        elif "right" in mode:
            self._attr["POS_X"] = 0
        elif "center" in mode:
            self._attr["POS_X"] = -rect.width() / 2

        self._attr["_OLD_POS"] = self.pos()



import os
import random
import json
import codecs
from collections import defaultdict

from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from qtpy.QtGui import QColor
from qtpy.QtGui import QFont
from qtpy.QtGui import QImage
from qtpy.QtGui import QPainter


from nezzle.graphics import NodeClassFactory
from nezzle.graphics import LinkClassFactory
from nezzle.graphics import LabelClassFactory
from nezzle.graphics import ArrowClassFactory
from nezzle.graphics import Network
from nezzle.constants import DEFAULT_SCENE_WIDTH, DEFAULT_SCENE_HEIGHT
from nezzle.utils import extract_name_and_ext

import math
from nezzle.utils.math import rotate, dist, internal_division


def read_metadata_from_sif(fpath):
    metadata = {}
    interactions = defaultdict(int)
    with codecs.open(fpath, "r", encoding="utf-8-sig") as fin:
        for i, line in enumerate(fin):
            if line.isspace():
                continue

            items = line.split()
            str_src, str_link_type, str_tgt = items[:3]
            interactions[str_link_type.strip()] += 1

    metadata["NETWORK_NAME"] = os.path.basename(fpath)
    metadata["INTERACTIONS"] = interactions
    return metadata


def read_metadata_from_json(fpath):
    metadata = {}
    interactions = defaultdict(int)
    with codecs.open(fpath, "r", encoding="utf-8") as fin:
        dict_net = json.loads(fin.read())

    for link in dict_net["LINKS"]:
        str_head_type = link["HEAD"]["TYPE"].title()
        interactions[str_head_type] += 1

    metadata["NETWORK_NAME"] = dict_net["NAME"]
    metadata["INTERACTIONS"] = interactions
    return metadata

def read_metadata(fpath):
    if not fpath:
        raise ValueError("Invalid file path: %s"%(fpath))

    file_name_ext = os.path.basename(fpath)
    fname, fext = os.path.splitext(file_name_ext)

    if file_name_ext.endswith('.sif'):
        return read_metadata_from_sif(fpath)
    elif file_name_ext.endswith('.json'):
        return read_metadata_from_json(fpath)

    else:
        raise ValueError("Unsupported file type: %s"%(fext))


def read_network(fpath, link_map=None):
    if not fpath:
        raise ValueError("Invalid file path: %s"%(fpath))
    file_name_ext = os.path.basename(fpath)
    fname, fext = os.path.splitext(file_name_ext)

    if file_name_ext.endswith('.sif'):
        return read_sif(fpath, link_map)
    elif file_name_ext.endswith('.json'):
        return read_json(fpath, link_map)

    else:
        raise ValueError("Unsupported file type: %s"%(fext))
# end of read_network


def read_sif(fpath, link_map=None):

    scene_width = DEFAULT_SCENE_WIDTH
    scene_height = DEFAULT_SCENE_HEIGHT

    fname = os.path.basename(fpath)
    net = Network(fname)
    #net.scene.setBackgroundBrush(QColor(0, 0, 0, 0))
    net.scene.setBackgroundBrush(Qt.white)  #(Qt.transparent)

    with codecs.open(fpath, "r", encoding="utf-8-sig") as fin:
        NodeClass = NodeClassFactory.create("ELLIPSE_NODE")
        nodes = {}
        counter_node = 0
        counter_link = 0
        for i, line in enumerate(fin):
            if line.isspace():
                continue

            items = line.split()
            str_src, str_link_type, str_tgt = items[:3]

            if link_map and (str_link_type not in link_map):
                raise ValueError("Undefined link type: %s"%(str_link_type))

            color = Qt.white
            half_width = scene_width/2
            half_height = scene_height/2

            range_x = (-scene_width/4, scene_width/4)
            range_y = (-scene_height/4, scene_height/4)

            sx = half_width + random.uniform(*range_x)
            sy = half_height + random.uniform(*range_y)

            tx = half_width + random.uniform(*range_x)
            ty = half_height + random.uniform(*range_y)

            width = 50
            height = 35

            if str_src in nodes:
                src = nodes[str_src]
            else:
                counter_node += 1
                src = NodeClass(str_src, width=width, height=height,
                                pos=QPointF(sx, sy))

                src["FILL_COLOR"] = color
                src['BORDER_COLOR'] = Qt.darkGray
                nodes[str_src] = src
            # end of else

            if str_tgt in nodes:
                tgt = nodes[str_tgt]
            else:
                counter_node += 1
                tgt = NodeClass(str_tgt, width=width, height=height,
                                pos=QPointF(tx, ty))

                tgt["FILL_COLOR"] = color
                tgt['BORDER_COLOR'] = Qt.darkGray
                nodes[str_tgt] = tgt
            # end of else

            counter_link += 1

            head = None
            ArrowClass = None

            if link_map:
                head_type = link_map[str_link_type]
                ArrowClass = ArrowClassFactory.create(head_type)

            if ArrowClass:
                head = ArrowClass()

            if str_src == str_tgt: # Self-loop link
                LinkClass = LinkClassFactory.create('SELFLOOP_LINK')
                iden = "%s%s%s" % (str_src, str_link_type, str_src)
                link = LinkClass(iden=iden,
                                 name=str_link_type,
                                 node=src,
                                 head=head)

                link["FILL_COLOR"] = QColor(100, 100, 100, 100)

            else:
                LinkClass = LinkClassFactory.create('CURVED_LINK')
                iden = "%s%s%s" % (str_src, str_link_type, str_tgt)
                link = LinkClass(iden=iden,
                                 name= str_link_type,
                                 source=src, target=tgt,
                                 head=head)

                link["FILL_COLOR"] = Qt.black

            src.add_link(link)
            tgt.add_link(link)
            net.add_link(link)
        # end of for: reading each line of SIF file

        # Add nodes and labels in network
        # font = QFont()
        # font.setFamily("Tahoma")
        # font.setPointSize(10)
        LabelClass = LabelClassFactory.create("TEXT_LABEL")
        for str_name, node in nodes.items():
            net.add_node(node)
            label = LabelClass(node, str_name)
            #label.font = font
            label["FONT_FAMILY"] = "Tahoma"
            label["FONT_SIZE"] = 10
            rect = label.boundingRect()
            label.setPos(-rect.width()/2, -rect.height()/2)
            net.add_label(label)
            nodes[str_name] = node

    # end of with

    for src, tgt, attr in net.nxdg.edges(data=True):
        if net.nxdg.has_edge(tgt, src):
            if src == tgt:  # Skip selfloops
                continue

            link = attr['GRAPHICS']
            mid = internal_division(link.pos_src, link.pos_tgt, 0.5, 0.5)
            d = dist(link.pos_src, mid)/math.cos(math.pi/4)
            cp = rotate(link.pos_src, mid, -30, d)
            link.ctrl_point.setPos(cp)

    return net
# end of def


def read_json(fpath, link_map):
    with codecs.open(fpath, "r", encoding="utf-8") as fin:
        dict_net = json.loads(fin.read())

        for link in dict_net["LINKS"]:
            head_type_ori = link["HEAD"]["TYPE"].title()
            head_type_new = link_map[head_type_ori]
            link["HEAD"]["TYPE"] = head_type_new.upper()

    return Network.from_dict(dict_net)


def write_network(net, fpath):
    file_name_ext = os.path.basename(fpath)
    file_name_ext = file_name_ext.casefold()

    if file_name_ext.endswith('.sif'):
        write_sif(net, fpath)
    elif file_name_ext.endswith('.json'):
        write_json(net, fpath)
    else:
        raise ValueError("Unsupported file type: %s" % (file_name_ext))


def write_sif(net, fpath):
    with codecs.open(fpath, "w", encoding="utf-8") as fout:
        for iden, link in net.links.items():
            if 'SELFLOOP' in link.ITEM_TYPE:
                args = (link.node.iden, link.name, link.node.iden)
            else:
                args = (link.source.iden, link.name, link.target.iden)
            fout.write("%s\t%s\t%s\n"%args)


def write_json(net, fpath):
    with codecs.open(fpath, "w", encoding="utf-8") as fout:
        fout.write(json.dumps(net.to_dict()))


def write_image(net,
                fpath,
                transparent=True,
                quality=100,
                scale_width=100, scale_height=100,
                dpi_width=350, dpi_height=350,
                pad_width=10, pad_height=10):

    fname, fext = extract_name_and_ext(fpath)

    scene = net.scene
    scene.clearSelection()
    scene.clearFocus()
    brect = scene.itemsBoundingRect()
    brect.adjust(-pad_width, -pad_height, +2*pad_width, +2*pad_height)

    image = QImage((scale_width/100.0) * brect.width(),
                   (scale_height/100.0) * brect.height(),
                   QImage.Format_ARGB32_Premultiplied)

    # [REF] http://stackoverflow.com/a/13425280/4136588
    # dpm = 300 / 0.0254 # ~300 DPI
    dpm_width = dpi_width / 0.0254
    dpm_height = dpi_height / 0.0254
    image.setDotsPerMeterX(dpm_width)
    image.setDotsPerMeterY(dpm_height)

    bbrush = scene.backgroundBrush()

    painter = QPainter(image)
    if not transparent or fext in ["jpeg", "jpg"]:
        image.fill(Qt.white)
        painter.setPen(Qt.NoPen)
        painter.setBrush(bbrush.color())
        painter.drawRect(0, 0, image.width(), image.height())
    elif fext in ['png']:
        image.fill(bbrush.color())

    painter.setRenderHints(QPainter.TextAntialiasing
                           | QPainter.Antialiasing
                           | QPainter.SmoothPixmapTransform
                           | QPainter.HighQualityAntialiasing)

    scene.render(painter)
    image.save(fpath, fext.upper(), quality)
    painter.end()





from typing import AnyStr
from typing import Union
from typing import Type

from nezzle.graphics.nodes.basenode import BaseNode
from nezzle.graphics.nodes.nodefactory import NodeClassFactory


class NodeConverter(object):

    @staticmethod
    def convert(node: BaseNode, node_type: Union[Type, AnyStr]):
        if isinstance(node_type, str):
            node_type = NodeClassFactory.create(node_type)

        if type(node) == node_type:
            return

        attr = node.to_dict()
        attr["ITEM_TYPE"] = node_type.ITEM_TYPE
        new_node = node_type.from_dict(attr=attr)
        return new_node




from qtpy.QtCore import Qt

from qtpy.QtWidgets import QDialog


from nezzle.ui.ui_progressdialog import Ui_ProgressDialog



class ProgressDialog(QDialog, Ui_ProgressDialog):

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.setModal(False)

        self.setWindowFlags(Qt.FramelessWindowHint)
        self.setGeometry(parent.width()/2,
                         parent.height()/2,
                         self.width(), self.height())

        self._progress_bar = self.ui_progressBar
        self._progress_bar.setMaximum(100)
        self._progress_bar.setValue(-1)


    def exec(self):
        self._progress_bar.setMaximum(0)
        super().exec()

    def show(self):
        self._progress_bar.setMaximum(0)
        #self.exc(self.nav, self.net)
        super().show()

    def close(self):
        self._progress_bar.setMaximum(100)
        super().close()


"""
Identifying sub-control points (sps), Q and D points
for quadratic Bezier curve with a width is based on
the following article:

"Quadratic bezier offsetting with selective subdivision"
by Gabriel Suchowolski, Jul 10, 2012

http://microbians.com/?page=math&id=math-quadraticbezieroffseting
"""

from typing import Union
import numpy as np
# from cachetools import cached

from qtpy.QtCore import QPointF

from nezzle.utils import dist
from nezzle.utils import dot
from nezzle.utils import length
from nezzle.utils import internal_division
from nezzle.utils import solve_cubic
from nezzle.utils import rotate


# def _integrate_sqrt_quad(coeff: np.ndarray, bounds: np.ndarray) -> float:
#     """
#     An exact solution of integrating sqrt(a*x^2 + b*x + c) over bounds.
#
#     Args:
#         coeff: A sequence of the coefficients, (a, b, c).
#         bounds: Upper nad lower bounds.
#     """
#
#     if np.abs(bounds[0] - bounds[1]) < 1e-5:
#         return 0.0
#
#     a, b, c = coeff
#
#     """
#     # The function of definite integral
#     # (1) Calcuate Fq
#     f = np.sqrt(a*q**2 + b*q + c)
#     g = (2*a*q + b) / (a**0.5) + 2*f
#     Fq = (b/(4.*a) + 0.5*q)*f+(4*a*c-b**2)/(8*a**(1.5))*np.log(g)
#
#     # (2) Calculate Fp
#     f = np.sqrt(a*p**2 + b*p + c)
#     g = (2 * a * p + b) / (a ** 0.5) + 2 * f
#     Fp = (b / (4. * a) + 0.5 * p) * f + (4 * a * c - b ** 2) / (8 * a ** (1.5)) * np.log(g)
#     """
#
#     f = np.sqrt(a * bounds ** 2 + b * bounds + c)
#     g = (2 * a * bounds + b) / (a ** 0.5) + 2 * f
#     F = (b / (4. * a) + 0.5 * bounds) * f + (4 * a * c - b ** 2) / (8 * a ** (1.5)) * np.log(g)
#     return F[1] - F[0]
#
#
# def arc_length(x: np.ndarray, y: np.ndarray, t1: float, t2: float) -> float:
#     """
#     The arc length of quadratic Bezier curve, given (t1, t2)
#
#     Args:
#         x: The x-axis coordinates of the three control points of quadratic Bezier curve.
#         y: The x-axis coordinates of the three control points of quadratic Bezier curve.
#         t1: The lower bound.
#         t2: The upper bound.
#
#     '''math
#     x'(t)^2 = 4*((P-Q)^2*t^2 -2*P*(P-Q)*t + P^2)
#     y'(t)^2 = 4*((R-S)^2*t^2 -2*R*(R-S)*t + R^2)
#     '''
#
#     Integrate 2*\sprt{x'(t)^2 + y'(t)^2} over (t1, t2)
#     """
#
#     #p0, p1, p2 = cps
#
#     p = x[1] - x[0]  #p1.x() - p0.x()
#     q = x[2] - x[1]  #p2.x() - p1.x()
#     r = y[1] - y[0]  #p1.y() - p0.y()
#     s = y[2] - y[1]  #p2.y() - p1.y()
#
#     coeff = np.zeros(3, dtype=np.float64)
#     coeff[0] = (p - q) ** 2 + (r - s) ** 2
#     coeff[1] = -2 * ((p - q) * p + (r - s) * r)
#     coeff[2] = p ** 2 + r ** 2
#
#     return 2 * _integrate_sqrt_quad(coeff, np.array([t1, t2]))

def _integrate_sqrt_quad(coeff: np.ndarray,
                         lower: Union[np.ndarray, float],
                         upper: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """
    An exact solution of integrating sqrt(a*x^2 + b*x + c) over bounds.

    Args:
        coeff: A sequence of the coefficients, (a, b, c).
        bounds: Upper nad lower bounds.
    """

    a, b, c = coeff


    # The function of definite integral
    # (1) Calcuate Fl (lower)
    f = np.sqrt(a*lower**2 + b*lower + c)
    g = (2*a*lower + b) / (a**0.5) + 2*f
    Fl = (b/(4.*a) + 0.5*lower)*f+(4*a*c-b**2)/(8*a**(1.5))*np.log(g)

    # (2) Calculate Fu (upper)
    f = np.sqrt(a*upper**2 + b*upper + c)
    g = (2 * a * upper + b) / (a ** 0.5) + 2 * f
    Fu = (b / (4. * a) + 0.5 * upper) * f + (4 * a * c - b ** 2) / (8 * a ** (1.5)) * np.log(g)

    return Fu - Fl


def arc_length(x: np.ndarray,
               y: np.ndarray,
               lower: Union[np.ndarray, float],
               upper: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """
    The arc length of quadratic Bezier curve, given (t1, t2)

    Args:
        x: The x-axis coordinates of the three control points of quadratic Bezier curve.
        y: The x-axis coordinates of the three control points of quadratic Bezier curve.
        lower: The lower bound.
        upper: The upper bound.

    '''math
    x'(t)^2 = 4*((P-Q)^2*t^2 -2*P*(P-Q)*t + P^2)
    y'(t)^2 = 4*((R-S)^2*t^2 -2*R*(R-S)*t + R^2)
    '''

    Integrate 2*\sprt{x'(t)^2 + y'(t)^2} over (t1, t2)
    """

    #p0, p1, p2 = cps

    p = x[1] - x[0]  #p1.x() - p0.x()
    q = x[2] - x[1]  #p2.x() - p1.x()
    r = y[1] - y[0]  #p1.y() - p0.y()
    s = y[2] - y[1]  #p2.y() - p1.y()

    coeff = np.zeros(3, dtype=np.float64)
    coeff[0] = (p - q) ** 2 + (r - s) ** 2
    coeff[1] = -2 * ((p - q) * p + (r - s) * r)
    coeff[2] = p ** 2 + r ** 2

    return 2 * _integrate_sqrt_quad(coeff, lower, upper)



def nearest_point(cps, pa):
    """
    Find the nearest point of the quadratic bezier curve
    from the control point

    cps: sequence of control points (QPointF)
    pa: arbitrary point
    """
    p1 = cps[0]
    pc = cps[1]
    p2 = cps[2]

    coords_x = 1e8*np.array([p1.x(), p2.x(), pc.x(), pa.x()])
    coords_y = 1e8*np.array([p1.y(), p2.y(), pc.y(), pa.y()])

    mean_x = np.average(coords_x)
    mean_y = np.average(coords_y)

    sv_x = np.std(coords_x)
    sv_y = np.std(coords_y)

    sv_x = sv_x if sv_x != 0 else 1.0
    sv_y = sv_y if sv_y != 0 else 1.0

    p1_scaled = QPointF((p1.x()-mean_x)/sv_x, (p1.y()-mean_y)/sv_y)
    p2_scaled = QPointF((p2.x()-mean_x)/sv_x, (p2.y()-mean_y)/sv_y)
    pc_scaled = QPointF((pc.x()-mean_x)/sv_x, (pc.y()-mean_y)/sv_y)
    pa_scaled = QPointF((pa.x()-mean_x)/sv_x, (pa.y()-mean_y)/sv_y)

    v0 = pc_scaled - p1_scaled
    v1 = p2_scaled - pc_scaled
    v3 = pa_scaled - p1_scaled

    diff_v1v0 = v1 - v0
    sq_diff_v1v0 = diff_v1v0.x()**2 + diff_v1v0.y()**2
    sq_v0 = v0.x()**2 + v0.y()**2

    a = sq_diff_v1v0  # dot((v1 - v0), (v1 - v0))
    b = 3*(dot(v1, v0) - sq_v0)  # 3*(dot(v1, v0) - dot(v0, v0))
    c = 2*sq_v0 - dot(v3, diff_v1v0)  # 3*dot(v0, v0) - dot(v1, v0)
    d = -dot(v3, v0)

    sols = solve_cubic(a, b, c, d)
    t = sols[0]
    f = a*t**3 + b*t**2 + c*t + d

    assert np.allclose(np.imag(t), 0)
    t = np.real(t)

    if t<0 or t>1:
        raise FloatingPointError("Abnormal value for parameter, t: %f"%(t))

    pt = (1-t)**2*p1 + 2*t*(1-t)*pc + t**2*p2
    return pt, t


def identify_sps(cps):
    pt0 = cps[0]
    pc = cps[1]
    pt4 = cps[2]

    pt2, t2 = nearest_point(cps, pc)

    # Control points of subdivision
    p11 = internal_division(pt0, pc, t2, 1 - t2)
    p12 = internal_division(pc, pt4, t2, 1 - t2)

    pt1, t1 = nearest_point((pt0, p11, pt2), p11)
    pt3, t3 = nearest_point((pt2, p12, pt4), p12)

    c01 = internal_division(pt0, p11, t1, 1 - t1)
    c12 = internal_division(p11, pt2, t1, 1 - t1)
    c23 = internal_division(pt2, p12, t3, 1 - t3)
    c34 = internal_division(p12, pt4, t3, 1 - t3)

    return [pt0, c01, pt1, c12, pt2, c23, pt3, c34, pt4]


def identify_qps(sps, width):

    qps_top = []
    qps_bottom = []

    for i in range(0, len(sps)-1, 2):
        p, cp = sps[i], sps[i+1]
        q_top = rotate(p, cp, -90, width/2)
        q_bottom = rotate(p, cp, 90, width/2)
        qps_top.append(q_top)
        qps_bottom.append(q_bottom)

    q_top = rotate(sps[-1], sps[-2], 90, width/2)
    q_bottom = rotate(sps[-1], sps[-2], -90, width/2)
    qps_top.append(q_top)
    qps_bottom.append(q_bottom)

    return qps_top, qps_bottom


def identify_dps(sps, width):
    dps_top = []
    dps_bottom =[]

    for i in range(1, len(sps), 2):
        p1, cp, p2 = sps[i-1], sps[i], sps[i+1]
        d_top = identify_dp(cp, p1, p2, width, 1)
        d_bottom = identify_dp(cp, p1, p2, width, -1)
        dps_top.append(d_top)
        dps_bottom.append(d_bottom)

    return dps_top, dps_bottom


def identify_dp(ap, p1, p2, width, angle_sign):
    m1 = rotate(ap, p1, angle_sign*90, 1.0) - ap
    m2 = rotate(ap, p2, -angle_sign*90, 1.0) - ap
    m = m1 + m2
    k = m*width/dot(m, m)
    return ap + k


from nezzle.graphics.links.straightlink import StraightLink
from nezzle.graphics.links.curvedlink import CurvedLink
from nezzle.graphics.links.elbowlink import VerticalElbowLink
from nezzle.graphics.links.elbowlink import HorizontalElbowLink
from nezzle.graphics.links.selflooplink import SelfloopLink


class LinkClassFactory(object):

    @staticmethod
    def create(link_type):
        if link_type.upper() == 'STRAIGHT_LINK':
            return StraightLink
        elif link_type.upper() == "VERTICAL_ELBOW_LINK":
            return VerticalElbowLink
        elif link_type.upper() == "HORIZONTAL_ELBOW_LINK":
            return HorizontalElbowLink
        elif link_type.upper() == 'CURVED_LINK':
            return CurvedLink
        elif link_type.upper() == 'SELFLOOP_LINK':
            return SelfloopLink
        else:
            raise TypeError("Undefined link type: %s" % (link_type))



from qtpy.QtCore import Qt
from qtpy.QtCore import QObject
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QMenu
from qtpy.QtWidgets import QWidget

from nezzle.widgets.console import ConsoleWidget


class ConsoleTabManager(QObject):
    def __init__(self, mw):
        super().__init__(parent=mw)

        self.mw = mw  # Main Window
        self._tab_widget = self.mw.ui_consoleTabWidget
        self._console_widget = ConsoleWidget(parent=self.tab_widget)

        self.tab_widget.addTab(self.console_widget, "Console")
        self.tab_widget.setTabsClosable(False)
        self.tab_widget.setMovable(True)

        #self.tab_widget.tabCloseRequested.connect(self.closeTab)

        # Create context menu
        self.tab_widget.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tab_widget.customContextMenuRequested.connect(self.on_context_menu)
        self.pop_menu = QMenu(self.tab_widget)
        self.action_restart_console = self.pop_menu.addAction('Restart console')
        #self.appendConsole()

    @property
    def tab_widget(self):
        return self._tab_widget

    @property
    def console_widget(self):
        return self._console_widget

    def restart_console(self):
        self.console_widget.reset()
        self.update_console_vars()

    def update_console_vars(self):
        net = None
        selected = None
        if self.mw.nt_manager.current_item:
            net = self.mw.nt_manager.current_item.data()
            selected = net.scene.selectedItems()

        vars = {'net': net,
                'console': self.console_widget,
                'view': self.mw.sv_manager.view,
                'nav': self.mw.nt_manager,
                'selected': selected}

        self.console_widget.pushVariables(vars)

    def on_context_menu(self, point):
        action = self.pop_menu.exec_(self.tab_widget.mapToGlobal(point))
        if action == self.action_restart_console:
            self.restart_console()

    # @Slot(int)
    # def closeTab(self, currentIndex):
    #     console_widget = self.widget(currentIndex)
    #     console_widget.stop()
    #     self.removeTab(currentIndex)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './nezzle/ui/ui_progressdialog.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ProgressDialog(object):
    def setupUi(self, ProgressDialog):
        ProgressDialog.setObjectName("ProgressDialog")
        ProgressDialog.resize(449, 40)
        ProgressDialog.setSizeGripEnabled(False)
        ProgressDialog.setModal(False)
        self.verticalLayout = QtWidgets.QVBoxLayout(ProgressDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.ui_progressBar = QtWidgets.QProgressBar(ProgressDialog)
        self.ui_progressBar.setProperty("value", 0)
        self.ui_progressBar.setTextVisible(False)
        self.ui_progressBar.setObjectName("ui_progressBar")
        self.verticalLayout.addWidget(self.ui_progressBar)

        self.retranslateUi(ProgressDialog)
        QtCore.QMetaObject.connectSlotsByName(ProgressDialog)

    def retranslateUi(self, ProgressDialog):
        _translate = QtCore.QCoreApplication.translate
        ProgressDialog.setWindowTitle(_translate("ProgressDialog", "Dialog"))




import math

from qtpy.QtCore import QRectF
from qtpy.QtGui import QPainterPath

from nezzle.graphics.links.baselink import BaseLink
from nezzle.graphics.mixins import Lockable


@Lockable
class SelfloopLink(BaseLink):

    ITEM_TYPE = 'SELFLOOP_LINK'

    def __init__(self, iden, node, *args, **kwargs):
        self._node = node
        self._radius_core = None
        self._angle_offset = None
        self._angle_sweep = None
        self._angle_begin = None

        super().__init__(iden, *args, **kwargs)

        self.node.add_link(self)

    @property
    def node(self):
        return self._node

    def initialize(self):
        self._create_path()

    def _identify_pos(self):
        node_hw = self._node.width/2  # Half of nodes width
        node_hh = self._node.height/2  # Half of nodes height
        len_minor = min(node_hw, node_hh)
        self._radius_core = (node_hw**2 + node_hh**2)/(2*len_minor)

        if node_hw > node_hh:  # X-axis is major.
            x = -node_hw
            y = -self._radius_core
        else:  # Y-axis is major.
            x = -self._radius_core
            y = -node_hh

        self.setPos(x+self._node.x(), y+self._node.y())

    def _calculate_angle_offset(self):
        if self.head:
            offset = self.head.offset + self.head.height
            self._angle_offset = math.degrees(offset/self._radius_core)
        else:
            self._angle_offset = 0

    def _calculate_angle_sweep(self):
        node_hw = self._node.width/2  # Half of nodes width
        node_hh = self._node.height/2  # Half of nodes height
        radius = self._radius_core
        self._angle_sweep = 360 - self._angle_offset

        # Law of cosine
        cosine = (2*radius**2-(node_hw**2+node_hh**2))/(2*radius**2)
        self._angle_sweep -= math.degrees(math.acos(cosine))

    def _calculate_head_angle(self):
        """Calculate the angle between the tangent line of
        the core circle at the axis between 3rd and 4th quadrants
        (where the head angle is not rotated yet) and the tangent
        line of head pos.
        """
        self._angle_head = self._angle_sweep - (self._angle_begin-90)

    def _calculate_angle_begin(self):
        node_hw = self._node.width / 2  # Half of nodes width
        #node_hh = self._node.height / 2  # Half of nodes height
        radius = self._radius_core

        self._angle_begin = 270

        """Law of cosine
           dist_sq is the square of distance between
           minus Y-axis of loop and ellipse.
        """
        dist_sq = (self.x()-self._node.x()+node_hw)**2 \
                   + (self.y()-self._node.y()+radius)**2

        cosine = (2*radius**2-dist_sq)/(2*radius**2)
        self._angle_begin += math.degrees(math.acos(cosine))

    def _identify_head(self):
        SelfloopLink._identify_head_pos(self)
        SelfloopLink._calculate_head_angle(self)
        super()._create_head_path()

    def _identify_head_pos(self):
        radius = self._radius_core
        rect_core = QRectF(-radius, -radius,
                           2*radius, 2*radius)

        pos = self._identify_arc_pos(rect_core,
                                     -self._angle_sweep)
        self.pos_head.setX(pos.x())
        self.pos_head.setY(pos.y())

    def _identify_arc_pos(self, rect, angle):
        path = QPainterPath()
        path.arcMoveTo(rect, self._angle_begin+angle)
        return path.currentPosition()

    def _create_path(self):

        self._identify_pos()

        # The calculating the angles, offset and sweep, should be done
        # in the sequence: (1) offset, (2) sweep.
        self._calculate_angle_offset()
        self._calculate_angle_sweep()
        self._calculate_angle_begin()

        """
        [REF] http://stackoverflow.com/a/17123051/4136588
        """
        self._path_paint = QPainterPath()
        path = self._path_paint

        radius = self._radius_core
        half_width = self.width/2
        diameter = 2*radius
        rect_outer = QRectF(-radius-half_width,
                            -radius-half_width,
                            diameter+self.width,
                            diameter+self.width)

        rect_inner = QRectF(-radius+half_width,
                            -radius+half_width,
                            diameter-self.width,
                            diameter-self.width)

        begin_angle = self._angle_begin  #270
        path.arcMoveTo(rect_outer, begin_angle)
        path.arcTo(rect_outer, begin_angle, -self._angle_sweep)

        if self.head:
            SelfloopLink._identify_head(self)
            self._path_paint.connectPath(self._path_head)
        else:
            path.lineTo(self._identify_arc_pos(rect_inner, -self._angle_sweep))

        path.arcTo(rect_inner, begin_angle-self._angle_sweep, self._angle_sweep)
        path.lineTo(self._identify_arc_pos(rect_outer, 0))
        path.closeSubpath()

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)

    def to_dict(self):
        attr = super().to_dict()
        attr['ID_NODE'] = self.node.iden
        return attr

    @classmethod
    def from_dict(cls, attr, node):
        iden = attr.pop('ID')
        width = attr.pop('WIDTH')

        obj = cls(iden, node, width=width)
        obj.head = cls.head_from_dict(attr)

        attr['ID_NODE'] = node.iden
        obj._attr.update(attr)
        return obj



from qtpy.QtCore import QPointF
from nezzle.utils import TriggerDict


class BaseArrow(object):

    TYPE = 'BASE_HEAD'

    def __init__(self, width, height, offset=4):

        self._attr = TriggerDict()
        self._attr['TYPE'] = self.TYPE

        self._offset = offset
        self._height = height
        self._width = width

        self._attr.set_trigger('WIDTH', self._trigger_set_width, when='set')
        self._attr.set_trigger('HEIGHT', self._trigger_set_height, when='set')
        self._attr.set_trigger('OFFSET', self._trigger_set_offset, when='set')
        
        self._attr['WIDTH'] = width
        self._attr['HEIGHT'] = height
        self._attr['OFFSET'] = offset

    # Read-write properties
    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, obj):
        self._parent = obj

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, val):
        self._attr['WIDTH'] = val
        self.update()

    def _trigger_set_width(self, key, value):
        self._width = value
        return value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, val):
        self._attr['HEIGHT'] = val
        self.update()

    def _trigger_set_height(self, key, value):
        self._height = value
        return value

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, val):
        if not hasattr(self, "_parent") or not self._parent:
            raise ValueError("A link should be assigned for this arrow before setting offset.")

        self._attr['OFFSET'] = val
        self.update()

    def _trigger_set_offset(self, key, value):
        self._offset = value
        return value

    # @property
    # def symbol(self):
    #     return self._symbol
    #
    # @symbol.setter
    # def symbol(self, val):
    #     self._symbol = val

    def update(self):
        self.parent.update()

    def identify_points(self, head, link_body_width, angle=None):
        raise NotImplementedError("identify_pos should be implemented!")

    def to_dict(self):
        dict_head = {}
        dict_head['TYPE'] = self.TYPE
        dict_head['WIDTH'] = self.width
        dict_head['HEIGHT'] = self.height
        dict_head['OFFSET'] = self.offset
        dict_head.update(self._attr)

        return dict_head

    @classmethod
    def from_dict(cls, dict_head):
        width = dict_head['WIDTH']
        height = dict_head['HEIGHT']
        offset = dict_head['OFFSET']
        #head_type = dict_head['TYPE']
        #ArrowClass = ArrowClassFactory.create(head_type)
        return cls(width, height, offset=offset)


class Triangle(BaseArrow):

    TYPE = "TRIANGLE"

    def __init__(self, width=10, height=10, *args, **kwargs):
        super().__init__(width, height, *args, **kwargs)

    def identify_points(self, head, link_body_width, transform=None):

        neck1 = head + QPointF(0, -link_body_width/2)
        neck2 = head + QPointF(0, +link_body_width/2)

        face1 = head + QPointF(0.0, -self.width/2)
        face2 = head + QPointF(0.0, +self.width/2)
        top = head + QPointF(self.height, 0)

        points = [neck1, face1, top, face2, neck2]

        # transform is a callable object, which defines its own transformation in __call__.
        if transform:
            for i, pt in enumerate(points):
                points[i] = transform(pt, head)

        return points
    # end of def identify_pos

    def set_size_from_link(self, link_width):
        self.width = 5*link_width
        self.height = 5*link_width
        self.parent.update()


class Hammer(BaseArrow):

    TYPE = "HAMMER"

    def __init__(self, width=14, height=2, *args, **kwargs):
        super().__init__(width, height, *args, **kwargs)

    def identify_points(self, head, link_body_width, transform=None):

        neck1 = head + QPointF(0, -link_body_width/2)
        neck2 = head + QPointF(0, +link_body_width/2)

        face1 = head + QPointF(0, -self.width/2)
        face2 = head + QPointF(self.height, -self.width/2)

        face3 = head + QPointF(self.height, +self.width/2)
        face4 = head + QPointF(0, +self.width/2)

        points = [neck1, face1, face2, face3, face4, neck2]

        if transform:
            for i, pt in enumerate(points):
                points[i] = transform(pt, head)

        return points
    # end of def identify_pos

    def set_size_from_link(self, link_width):
        self.width = 7*link_width
        self.height = link_width
        self.parent.update()


import numpy as np

from qtpy.QtCore import QPoint, QPointF, Qt
from qtpy.QtCore import QRectF
from qtpy.QtCore import QLineF

from qtpy.QtGui import QTransform
from qtpy.QtGui import QColor, QPainterPath, QPainter
from qtpy.QtGui import QPolygonF
from qtpy.QtGui import QPen
from qtpy.QtWidgets import QGraphicsScene, QGraphicsView
from qtpy.QtWidgets import QGraphicsItem

from nezzle.utils import dist
from nezzle.utils import dot
from nezzle.utils import length
from nezzle.utils import internal_division
from nezzle.utils import rotate
from nezzle.graphics import quadbezier
from nezzle.graphics.mixins import Lockable
from nezzle.graphics.baseitem import PainterOptionItem
from nezzle.graphics import ArrowClassFactory
from nezzle.graphics.nodes.basenode import BaseNode


class BaseLink(PainterOptionItem):
    __doc__ = r"""
       [`TwoNodeLink`][nezzle.graphics.links.baselink.TwoNodeLink]
       Applies a 2D convolution over an input signal composed of several input
       planes.
       In the simplest case, the output value of the layer with input size
       $`(N, C_{\text{in}}, H, W)`$ 
       and output $`(N, C_{\text{out}}, H_{\text{out}}, W_{\text{out}})`$ 
       can be precisely described as:
       
       ```math
           \text{out}(N_i, C_{\text{out}_j}) = \text{bias}(C_{\text{out}_j}) +
           \sum_{k = 0}^{C_{\text{in}} - 1} \text{weight}(C_{\text{out}_j}, k) \star \text{input}(N_i, k)
           
       ```
       where $`\star`$ is the valid 2D `cross-correlation`_ operator,
       ```math
       `N` is a batch size, `C` denotes a number of channels,
       `H` is a height of input planes in pixels, and `W` is
       width in pixels.
       ```

       This module supports `TensorFloat32<tf32_on_ampere>`.
       * ``stride`` controls the stride for the cross-correlation, a single
         number or a tuple.
       * :attr:`padding` controls the amount of padding applied to the input. It
         can be either a string {{'valid', 'same'}} or a tuple of ints giving the
         amount of implicit padding applied on both sides.
       * :attr:`dilation` controls the spacing between the kernel points; also
         known as the à trous algorithm. It is harder to describe, but this `link`_
         has a nice visualization of what :attr:`dilation` does.
       {groups_note}
       The parameters ``kernel_size``, `stride`, :attr:`padding`, :attr:`dilation` can either be:
           - a single ``int`` -- in which case the same value is used for the height and width dimension
           - a ``tuple`` of two ints -- in which case, the first `int` is used for the height dimension,
             and the second `int` for the width dimension
       Note:
           {depthwise_separable_note}
       Note:
           {cudnn_reproducibility_note}
       Note:
           ``padding='valid'`` is the same as no padding. ``padding='same'`` pads
           the input so the output has the shape as the input. However, this mode
           doesn't support any stride values other than 1.
       Args:
           in_channels (int): Number of channels in the input image
           out_channels (int): Number of channels produced by the convolution
           kernel_size (int or tuple): Size of the convolving kernel
           stride (int or tuple, optional): Stride of the convolution. Default: 1
           padding (int, tuple or str, optional): Padding added to all four sides of
               the input. Default: 0
           padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
               ``'replicate'`` or ``'circular'``. Default: ``'zeros'``
           dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
           groups (int, optional): Number of blocked connections from input
               channels to output channels. Default: 1
           bias (bool, optional): If ``True``, adds a learnable bias to the
               output. Default: ``True``
       """  + r"""
       Shape:
           - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
           - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where
             .. math::
                 H_{out} = \left\lfloor\frac{H_{in}  + 2 \times \text{padding}[0] - \text{dilation}[0]
                           \times (\text{kernel\_size}[0] - 1) - 1}{\text{stride}[0]} + 1\right\rfloor
             .. math::
                 W_{out} = \left\lfloor\frac{W_{in}  + 2 \times \text{padding}[1] - \text{dilation}[1]
                           \times (\text{kernel\_size}[1] - 1) - 1}{\text{stride}[1]} + 1\right\rfloor
       Attributes:
           weight (Tensor): the learnable weights of the module of shape
               :math:`(\text{out\_channels}, \frac{\text{in\_channels}}{\text{groups}},`
               :math:`\text{kernel\_size[0]}, \text{kernel\_size[1]})`.
               The values of these weights are sampled from
               :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
               :math:`k = \frac{groups}{C_\text{in} * \prod_{i=0}^{1}\text{kernel\_size}[i]}`
           bias (Tensor):   the learnable bias of the module of shape
               (out_channels). If :attr:`bias` is ``True``,
               then the values of these weights are
               sampled from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
               :math:`k = \frac{groups}{C_\text{in} * \prod_{i=0}^{1}\text{kernel\_size}[i]}`
       Examples:
           >>> # With square kernels and equal stride
           >>> m = nn.Conv2d(16, 33, 3, stride=2)
           >>> # non-square kernels and unequal stride and with padding
           >>> m = nn.Conv2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2))
           >>> # non-square kernels and unequal stride and with padding and dilation
           >>> m = nn.Conv2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2), dilation=(3, 1))
           >>> input = torch.randn(20, 16, 50, 100)
           >>> output = m(input)
       .. _cross-correlation:
           https://en.wikipedia.org/wiki/Cross-correlation
       .. _link:
           https://github.com/vdumoulin/conv_arithmetic/blob/master/README.md
       """

    """
    The very base class of all links
    """

    ITEM_TYPE = 'LINK'

    def __init__(self,
                 iden,
                 head=None,
                 width=2,
                 *args,
                 **kwargs):

        """

        """
        super().__init__(iden, *args, **kwargs)

        self._pos_head = QPointF()
        self._path_paint = QPainterPath()
        self._bounding_rect = QRectF()

        self.setFlags(QGraphicsItem.ItemIsSelectable
                      | QGraphicsItem.ItemIsFocusable
                      | QGraphicsItem.ItemSendsGeometryChanges)

        self._attr.set_trigger('WIDTH', self._trigger_set_width, when='set')
        self._attr.set('WIDTH', width, trigger=False)
        self._width = width

        self._attr.set_trigger('HEAD', self._trigger_set_head, when='set')
        self._attr.set('HEAD', head, trigger=False)

        self._head = head
        if head:
            self._head.parent = self

        self.initialize()

    def __str__(self):
        return 'Link(%s)'%(self._iden)

    @property
    def width(self):
        return self._attr['WIDTH']

    @width.setter
    def width(self, val):
        self._attr['WIDTH'] = val

    def set_width(self, val, head=True):
        if head and self.head:
            self.head.set_size_from_link(link_width=val)

        self._width = val
        self._attr.set('WIDTH', val, trigger=False)
        self.update()

    def _trigger_set_width(self, key, value):
        self.set_width(value)
        return value

    def _trigger_set_head(self, key, value):
        self._head = value
        if self._head:
            self._head.parent = self

        self._attr.set('HEAD', value, trigger=False)
        self.update()

        return value

    @property
    def head(self):
        return self._attr['HEAD']

    @head.setter
    def head(self, obj):
        self._attr['HEAD'] = obj

    @property
    def pos_head(self):
        return self._pos_head

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            """
            The following check prevents this item being selected
            simply according to boundingRect.
            """

            pos = event.pos() - self.pos()
            if self._path_paint.contains(pos):
                self.setSelected(True)
                event.accept()
        else:
            event.reject()


    """
    [Reference] https://stackoverflow.com/a/34199293
    QGraphicsItem::shape is used for object collision detection, hit tests and knowing where mouse clicks occur.
    In contrast, QGraphicsItem::boundingRect is used when drawing an object, knowing when an object is off the screen, or obscured.
    """

    def boundingRect(self):
        return self._bounding_rect

    def shape(self):
        return self._path_paint

    def itemChange(self, change, value):
        return super().itemChange(change, value)

    def _create_head_path(self):
        points = self.head.identify_points(self.pos_head,
                                             self.width,
                                             self._head_transform)
        path = QPainterPath()
        path.moveTo(points[0])
        for pt in points[1:]:
            path.lineTo(pt)

        self._path_head = path

    def update(self, *args, **kwargs):
        self._create_path()
        self._update_bounding_rect()
        self._invalidate()
        return super().update(*args, **kwargs)

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)
        painter.drawPath(self._path_paint)

    def to_dict(self):
        attr = super().to_dict()
        if self.head:
            attr['HEAD'] = self.head.to_dict()

        return attr

    @classmethod
    def head_from_dict(cls, attr):
        if 'HEAD' in attr:
            attr_head = attr.pop('HEAD')
            if attr_head:
                ArrowClass = ArrowClassFactory.create(attr_head['TYPE'])
                head = ArrowClass.from_dict(attr_head)
                return head
        return None


    @classmethod
    def from_dict(cls, attr):
        raise NotImplementedError()

    def copy(self):
        return self.from_dict(self.to_dict())

    def initialize(self):
        pass

    def _identify_pos(self):
        raise NotImplementedError()

    def _create_path(self):
        raise NotImplementedError()

    def _identify_head(self):
        raise NotImplementedError()

    def is_movable(self):
        return False

    def _update_bounding_rect(self):
        rect = self._path_paint.boundingRect()
        if self._pen:
            pad = 2 * self._pen.width()
            rect.adjust(-pad, -pad, +pad, +pad)

        self._bounding_rect = rect


class TwoNodeLink(BaseLink):

    ITEM_TYPE = 'TWO_NODE_LINK'

    def __init__(self, iden, source, target, *args, **kwargs):
        self._source = source
        self._target = target
        self.source.add_link(self)
        self.target.add_link(self)
        super().__init__(iden, *args, **kwargs)

    def __str__(self):
        str_link_type = ''.join([word.title() for word in self.ITEM_TYPE.split('_')])
        return "%s(%s, %s)"%(str_link_type, self.source.name, self.target.name)

    def __eq__(self, other):
        return id(other) == id(self)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, obj: BaseNode):
        if not obj.has_link(self):
            obj.add_link(self)
        self._source = obj

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, obj: BaseNode):
        if not obj.has_link(self):
            obj.add_link(self)
        self._target = obj

    # Read-only properties
    @property
    def pos_src(self):
        """The position of source relative to its parent link
        """
        return self._source.pos() - self.pos()

    @property
    def pos_tgt(self):
        """The position of target relative to its parent link
        """
        return self._target.pos() - self.pos()

    def is_node_selected(self):
        return self.source.isSelected() or self.target.isSelected()

    def are_nodes_close(self):
        """Decide whether the two nodes are overlapped to show the graphics of link appropriately.
        """
        v = self.pos_tgt - self.pos_src
        len_v = length(v)
        if len_v <= 1e-6:
            return True

        angle_rad = np.arccos(v.x() / len_v)
        radius_src = self.source.calculate_radius(angle_rad)

        v *= -1
        angle_rad = np.arccos(v.x() / len_v)
        radius_tgt = self.target.calculate_radius(angle_rad)

        return dist(self.pos_src, self.pos_tgt) < radius_src + radius_tgt

    def _identify_pos(self):
        m = internal_division(self.source.pos(), self.target.pos(), 0.5, 0.5)
        self.setPos(m)

    def to_dict(self):
        attr = super().to_dict()
        attr['ID_SOURCE'] = self.source.iden
        attr['ID_TARGET'] = self.target.iden
        return attr

    @classmethod
    def from_dict(cls, attr, source, target):
        iden = attr.pop('ID')
        width = attr.pop('WIDTH')

        obj = cls(iden, source, target, width=width)
        obj.head = cls.head_from_dict(attr)

        attr['ID_SOURCE'] = source.iden
        attr['ID_TARGET'] = target.iden

        obj._attr.update(attr)

        return obj

    def copy(self, source=None, target=None):
        if not source:
            source = self.source

        if not target:
            target = self.target

        return self.from_dict(self.to_dict(), source, target)



from nezzle.graphics.nodes.ellipsenode import EllipseNode
from nezzle.graphics.nodes.ellipsenode import CircleNode
from nezzle.graphics.nodes.rectanglenode import RectangleNode
from nezzle.graphics.nodes.rectanglenode import SquareNode


class NodeClassFactory(object):
    @staticmethod
    def create(node_type):
        if node_type.upper() == "ELLIPSE_NODE":
            return EllipseNode
        elif node_type.upper() == "CIRCLE_NODE":
            return CircleNode
        elif node_type.upper() == "RECTANGLE_NODE":
            return RectangleNode
        elif node_type.upper() == "SQUARE_NODE":
            return SquareNode
        else:
            raise TypeError("Undefined nodes type: %s" % (node_type))



# import warnings
# warnings.filterwarnings("ignore")

import os
from os.path import join as pjoin
import sys
import argparse

from qtpy import QtCore

from qtpy.QtWidgets import QApplication
from qtpy.QtWidgets import QStyleFactory
from qtpy.QtGui import QIcon

from nezzle.mainwindow import MainWindow

base_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(base_dir)


def main():
    argparser = argparse.ArgumentParser(description="Nezzle arguments for CUI execution")
    argparser.add_argument('-fc', '--fpath-code',
                           action='store',
                           dest='fpath_code',
                           default=None,
                           help="Designate the absolute file path of a source code file.")
    args = argparser.parse_args()

    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

    libpaths = QApplication.libraryPaths()
    libpaths.append(os.getcwd())
    QApplication.setLibraryPaths(libpaths)
    QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)

    app = QApplication(sys.argv)

    # It is possible to choose the theme of GUI
    # QApplication.setStyle(QStyleFactory.create("plastique"))

    mw = MainWindow()
    mw.setWindowTitle("Nezzle")
    droot_resources = pjoin(os.path.dirname(__file__), "resources")
    icon = QIcon(pjoin(droot_resources, "icon.png"))
    mw.setWindowIcon(icon)

    # Before showing the main window, process arguments.
    if args:
        if not args.fpath_code:
            pass
        elif os.path.isfile(args.fpath_code):
                mw.code_manager.process_code(args.fpath_code)
                mw.code_manager.file_path_edit.setText(args.fpath_code)
        else:
            sys.stderr.write("File path of source code is not valid: %s\n"%(args.fpath_code))

    # Show the main window.
    mw.show()

    sys.exit(app.exec())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './nezzle/ui/ui_mainwindow.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1347, 833)
        MainWindow.setDockNestingEnabled(True)
        self.ui_centralWidget = QtWidgets.QWidget(MainWindow)
        self.ui_centralWidget.setObjectName("ui_centralWidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.ui_centralWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.ui_mainFrame = QtWidgets.QFrame(self.ui_centralWidget)
        self.ui_mainFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.ui_mainFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.ui_mainFrame.setObjectName("ui_mainFrame")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.ui_mainFrame)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.ui_mainTabWidget = QtWidgets.QTabWidget(self.ui_mainFrame)
        self.ui_mainTabWidget.setMinimumSize(QtCore.QSize(0, 0))
        self.ui_mainTabWidget.setBaseSize(QtCore.QSize(0, 0))
        self.ui_mainTabWidget.setTabsClosable(False)
        self.ui_mainTabWidget.setMovable(True)
        self.ui_mainTabWidget.setObjectName("ui_mainTabWidget")
        self.verticalLayout_5.addWidget(self.ui_mainTabWidget)
        self.verticalLayout_2.addWidget(self.ui_mainFrame)
        MainWindow.setCentralWidget(self.ui_centralWidget)
        self.ui_menubar = QtWidgets.QMenuBar(MainWindow)
        self.ui_menubar.setGeometry(QtCore.QRect(0, 0, 1347, 36))
        self.ui_menubar.setObjectName("ui_menubar")
        self.ui_menuFile = QtWidgets.QMenu(self.ui_menubar)
        self.ui_menuFile.setObjectName("ui_menuFile")
        self.ui_menuView = QtWidgets.QMenu(self.ui_menubar)
        self.ui_menuView.setObjectName("ui_menuView")
        self.ui_menuEdit = QtWidgets.QMenu(self.ui_menubar)
        self.ui_menuEdit.setObjectName("ui_menuEdit")
        self.ui_menuGraphics = QtWidgets.QMenu(self.ui_menubar)
        self.ui_menuGraphics.setObjectName("ui_menuGraphics")
        self.ui_menuLock = QtWidgets.QMenu(self.ui_menuGraphics)
        self.ui_menuLock.setObjectName("ui_menuLock")
        self.ui_menuConvert = QtWidgets.QMenu(self.ui_menuGraphics)
        self.ui_menuConvert.setObjectName("ui_menuConvert")
        self.ui_menuLayout = QtWidgets.QMenu(self.ui_menubar)
        self.ui_menuLayout.setObjectName("ui_menuLayout")
        self.ui_menuAlign = QtWidgets.QMenu(self.ui_menuLayout)
        self.ui_menuAlign.setObjectName("ui_menuAlign")
        MainWindow.setMenuBar(self.ui_menubar)
        self.ui_statusbar = QtWidgets.QStatusBar(MainWindow)
        self.ui_statusbar.setObjectName("ui_statusbar")
        MainWindow.setStatusBar(self.ui_statusbar)
        self.ui_navigationDock = DockWidget(MainWindow)
        self.ui_navigationDock.setStyleSheet("")
        self.ui_navigationDock.setObjectName("ui_navigationDock")
        self.ui_controlDockContents = QtWidgets.QWidget()
        self.ui_controlDockContents.setObjectName("ui_controlDockContents")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.ui_controlDockContents)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.ui_groupNavigation = QtWidgets.QGroupBox(self.ui_controlDockContents)
        self.ui_groupNavigation.setObjectName("ui_groupNavigation")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.ui_groupNavigation)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.ui_navigationTree = NetworkTreeView(self.ui_groupNavigation)
        self.ui_navigationTree.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.ui_navigationTree.setObjectName("ui_navigationTree")
        self.verticalLayout_4.addWidget(self.ui_navigationTree)
        self.verticalLayout_3.addWidget(self.ui_groupNavigation)
        self.ui_groupCode = QtWidgets.QGroupBox(self.ui_controlDockContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_groupCode.sizePolicy().hasHeightForWidth())
        self.ui_groupCode.setSizePolicy(sizePolicy)
        self.ui_groupCode.setMaximumSize(QtCore.QSize(16777215, 130))
        self.ui_groupCode.setObjectName("ui_groupCode")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.ui_groupCode)
        self.verticalLayout_6.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.widget = QtWidgets.QWidget(self.ui_groupCode)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setObjectName("widget")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.widget)
        self.verticalLayout_9.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.verticalLayout_8.setContentsMargins(-1, 0, -1, -1)
        self.verticalLayout_8.setSpacing(10)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(self.widget)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.ui_codeFilePathEdit = QtWidgets.QLineEdit(self.widget)
        self.ui_codeFilePathEdit.setObjectName("ui_codeFilePathEdit")
        self.horizontalLayout_2.addWidget(self.ui_codeFilePathEdit)
        self.verticalLayout_8.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.ui_openCodeButton = QtWidgets.QPushButton(self.widget)
        self.ui_openCodeButton.setObjectName("ui_openCodeButton")
        self.horizontalLayout_3.addWidget(self.ui_openCodeButton)
        spacerItem1 = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.ui_runButton = QtWidgets.QPushButton(self.widget)
        self.ui_runButton.setObjectName("ui_runButton")
        self.horizontalLayout_3.addWidget(self.ui_runButton)
        spacerItem2 = QtWidgets.QSpacerItem(2, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.verticalLayout_8.addLayout(self.horizontalLayout_3)
        self.verticalLayout_9.addLayout(self.verticalLayout_8)
        self.verticalLayout_6.addWidget(self.widget)
        self.verticalLayout_3.addWidget(self.ui_groupCode)
        self.ui_navigationDock.setWidget(self.ui_controlDockContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.ui_navigationDock)
        self.ui_consoleDock = DockWidget(MainWindow)
        self.ui_consoleDock.setFeatures(QtWidgets.QDockWidget.AllDockWidgetFeatures)
        self.ui_consoleDock.setObjectName("ui_consoleDock")
        self.ui_consoleDockConetents = QtWidgets.QWidget()
        self.ui_consoleDockConetents.setObjectName("ui_consoleDockConetents")
        self.verticalLayout_10 = QtWidgets.QVBoxLayout(self.ui_consoleDockConetents)
        self.verticalLayout_10.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.ui_consoleTabWidget = QtWidgets.QTabWidget(self.ui_consoleDockConetents)
        self.ui_consoleTabWidget.setObjectName("ui_consoleTabWidget")
        self.verticalLayout_10.addWidget(self.ui_consoleTabWidget)
        self.ui_consoleDock.setWidget(self.ui_consoleDockConetents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.ui_consoleDock)
        self.ui_visStyleDock = QtWidgets.QDockWidget(MainWindow)
        self.ui_visStyleDock.setEnabled(False)
        self.ui_visStyleDock.setObjectName("ui_visStyleDock")
        self.ui_visStyleDockLayout = QtWidgets.QWidget()
        self.ui_visStyleDockLayout.setObjectName("ui_visStyleDockLayout")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.ui_visStyleDockLayout)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.ui_styleTabWidget = QtWidgets.QTabWidget(self.ui_visStyleDockLayout)
        self.ui_styleTabWidget.setTabPosition(QtWidgets.QTabWidget.North)
        self.ui_styleTabWidget.setMovable(True)
        self.ui_styleTabWidget.setObjectName("ui_styleTabWidget")
        self.ui_nodeTabLayout = QtWidgets.QWidget()
        self.ui_nodeTabLayout.setObjectName("ui_nodeTabLayout")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.ui_nodeTabLayout)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.ui_nodeStyleTable = QtWidgets.QTableView(self.ui_nodeTabLayout)
        self.ui_nodeStyleTable.setEnabled(False)
        self.ui_nodeStyleTable.setObjectName("ui_nodeStyleTable")
        self.horizontalLayout_5.addWidget(self.ui_nodeStyleTable)
        self.ui_styleTabWidget.addTab(self.ui_nodeTabLayout, "")
        self.ui_linkTabLayout = QtWidgets.QWidget()
        self.ui_linkTabLayout.setObjectName("ui_linkTabLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.ui_linkTabLayout)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.ui_linkStyleTable = QtWidgets.QTableView(self.ui_linkTabLayout)
        self.ui_linkStyleTable.setObjectName("ui_linkStyleTable")
        self.horizontalLayout_6.addWidget(self.ui_linkStyleTable)
        self.ui_styleTabWidget.addTab(self.ui_linkTabLayout, "")
        self.horizontalLayout_4.addWidget(self.ui_styleTabWidget)
        self.ui_visStyleDock.setWidget(self.ui_visStyleDockLayout)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(1), self.ui_visStyleDock)
        self.ui_historyDock = DockWidget(MainWindow)
        self.ui_historyDock.setFloating(False)
        self.ui_historyDock.setObjectName("ui_historyDock")
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.ui_historyView = HistoryView(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_historyView.sizePolicy().hasHeightForWidth())
        self.ui_historyView.setSizePolicy(sizePolicy)
        self.ui_historyView.setObjectName("ui_historyView")
        self.verticalLayout.addWidget(self.ui_historyView)
        self.ui_historyDock.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.ui_historyDock)
        self.ui_actionSaveNetwork = QtWidgets.QAction(MainWindow)
        self.ui_actionSaveNetwork.setObjectName("ui_actionSaveNetwork")
        self.ui_actionExit = QtWidgets.QAction(MainWindow)
        self.ui_actionExit.setObjectName("ui_actionExit")
        self.ui_actionOpenNetwork = QtWidgets.QAction(MainWindow)
        self.ui_actionOpenNetwork.setObjectName("ui_actionOpenNetwork")
        self.ui_actionNewView = QtWidgets.QAction(MainWindow)
        self.ui_actionNewView.setObjectName("ui_actionNewView")
        self.ui_actionSaveSimulation = QtWidgets.QAction(MainWindow)
        self.ui_actionSaveSimulation.setObjectName("ui_actionSaveSimulation")
        self.ui_actionOpenSimulation = QtWidgets.QAction(MainWindow)
        self.ui_actionOpenSimulation.setObjectName("ui_actionOpenSimulation")
        self.ui_actionViewNavigationDock = QtWidgets.QAction(MainWindow)
        self.ui_actionViewNavigationDock.setCheckable(True)
        self.ui_actionViewNavigationDock.setChecked(True)
        self.ui_actionViewNavigationDock.setObjectName("ui_actionViewNavigationDock")
        self.ui_actionViewConsoleDock = QtWidgets.QAction(MainWindow)
        self.ui_actionViewConsoleDock.setCheckable(True)
        self.ui_actionViewConsoleDock.setChecked(True)
        self.ui_actionViewConsoleDock.setObjectName("ui_actionViewConsoleDock")
        self.ui_actionViewVisualStyleDock = QtWidgets.QAction(MainWindow)
        self.ui_actionViewVisualStyleDock.setCheckable(True)
        self.ui_actionViewVisualStyleDock.setChecked(True)
        self.ui_actionViewVisualStyleDock.setObjectName("ui_actionViewVisualStyleDock")
        self.ui_actionExportImage = QtWidgets.QAction(MainWindow)
        self.ui_actionExportImage.setObjectName("ui_actionExportImage")
        self.ui_actionCopy = QtWidgets.QAction(MainWindow)
        self.ui_actionCopy.setObjectName("ui_actionCopy")
        self.ui_actionPaste = QtWidgets.QAction(MainWindow)
        self.ui_actionPaste.setObjectName("ui_actionPaste")
        self.actionLock_Nodes = QtWidgets.QAction(MainWindow)
        self.actionLock_Nodes.setObjectName("actionLock_Nodes")
        self.actionLock_Links = QtWidgets.QAction(MainWindow)
        self.actionLock_Links.setObjectName("actionLock_Links")
        self.ui_actionLockNodes = QtWidgets.QAction(MainWindow)
        self.ui_actionLockNodes.setCheckable(True)
        self.ui_actionLockNodes.setObjectName("ui_actionLockNodes")
        self.ui_actionLockLinks = QtWidgets.QAction(MainWindow)
        self.ui_actionLockLinks.setCheckable(True)
        self.ui_actionLockLinks.setObjectName("ui_actionLockLinks")
        self.ui_actionLockLabels = QtWidgets.QAction(MainWindow)
        self.ui_actionLockLabels.setCheckable(True)
        self.ui_actionLockLabels.setObjectName("ui_actionLockLabels")
        self.ui_actionAlignLeft = QtWidgets.QAction(MainWindow)
        self.ui_actionAlignLeft.setObjectName("ui_actionAlignLeft")
        self.ui_actionAlignCenter = QtWidgets.QAction(MainWindow)
        self.ui_actionAlignCenter.setObjectName("ui_actionAlignCenter")
        self.ui_actionAlignRight = QtWidgets.QAction(MainWindow)
        self.ui_actionAlignRight.setObjectName("ui_actionAlignRight")
        self.ui_actionAlignTop = QtWidgets.QAction(MainWindow)
        self.ui_actionAlignTop.setObjectName("ui_actionAlignTop")
        self.ui_actionAlignMiddle = QtWidgets.QAction(MainWindow)
        self.ui_actionAlignMiddle.setObjectName("ui_actionAlignMiddle")
        self.ui_actionAlignBottom = QtWidgets.QAction(MainWindow)
        self.ui_actionAlignBottom.setObjectName("ui_actionAlignBottom")
        self.ui_actionDistributeHorizontally = QtWidgets.QAction(MainWindow)
        self.ui_actionDistributeHorizontally.setObjectName("ui_actionDistributeHorizontally")
        self.ui_actionDistributeVertically = QtWidgets.QAction(MainWindow)
        self.ui_actionDistributeVertically.setObjectName("ui_actionDistributeVertically")
        self.ui_actionSelectAll = QtWidgets.QAction(MainWindow)
        self.ui_actionSelectAll.setObjectName("ui_actionSelectAll")
        self.ui_actionToCurvedLink = QtWidgets.QAction(MainWindow)
        self.ui_actionToCurvedLink.setObjectName("ui_actionToCurvedLink")
        self.ui_actionToStraightLink = QtWidgets.QAction(MainWindow)
        self.ui_actionToStraightLink.setObjectName("ui_actionToStraightLink")
        self.ui_actionToVerticalElbowLink = QtWidgets.QAction(MainWindow)
        self.ui_actionToVerticalElbowLink.setObjectName("ui_actionToVerticalElbowLink")
        self.ui_actionToHorizontalElbowLink = QtWidgets.QAction(MainWindow)
        self.ui_actionToHorizontalElbowLink.setObjectName("ui_actionToHorizontalElbowLink")
        self.ui_actionToEllipseNode = QtWidgets.QAction(MainWindow)
        self.ui_actionToEllipseNode.setObjectName("ui_actionToEllipseNode")
        self.ui_actionToRectangleNode = QtWidgets.QAction(MainWindow)
        self.ui_actionToRectangleNode.setObjectName("ui_actionToRectangleNode")
        self.ui_actionUndo = QtWidgets.QAction(MainWindow)
        self.ui_actionUndo.setShortcutContext(QtCore.Qt.WindowShortcut)
        self.ui_actionUndo.setObjectName("ui_actionUndo")
        self.ui_actionRedo = QtWidgets.QAction(MainWindow)
        self.ui_actionRedo.setObjectName("ui_actionRedo")
        self.ui_actionViewHistoryDock = QtWidgets.QAction(MainWindow)
        self.ui_actionViewHistoryDock.setCheckable(True)
        self.ui_actionViewHistoryDock.setChecked(True)
        self.ui_actionViewHistoryDock.setObjectName("ui_actionViewHistoryDock")
        self.ui_menuFile.addAction(self.ui_actionOpenNetwork)
        self.ui_menuFile.addSeparator()
        self.ui_menuFile.addAction(self.ui_actionSaveNetwork)
        self.ui_menuFile.addSeparator()
        self.ui_menuFile.addAction(self.ui_actionExportImage)
        self.ui_menuFile.addSeparator()
        self.ui_menuFile.addAction(self.ui_actionExit)
        self.ui_menuView.addAction(self.ui_actionViewNavigationDock)
        self.ui_menuView.addAction(self.ui_actionViewConsoleDock)
        self.ui_menuView.addAction(self.ui_actionViewHistoryDock)
        self.ui_menuEdit.addAction(self.ui_actionUndo)
        self.ui_menuEdit.addAction(self.ui_actionRedo)
        self.ui_menuEdit.addSeparator()
        self.ui_menuEdit.addAction(self.ui_actionCopy)
        self.ui_menuEdit.addAction(self.ui_actionPaste)
        self.ui_menuLock.addAction(self.ui_actionLockNodes)
        self.ui_menuLock.addAction(self.ui_actionLockLinks)
        self.ui_menuLock.addAction(self.ui_actionLockLabels)
        self.ui_menuConvert.addAction(self.ui_actionToEllipseNode)
        self.ui_menuConvert.addAction(self.ui_actionToRectangleNode)
        self.ui_menuConvert.addSeparator()
        self.ui_menuConvert.addAction(self.ui_actionToStraightLink)
        self.ui_menuConvert.addAction(self.ui_actionToCurvedLink)
        self.ui_menuConvert.addAction(self.ui_actionToVerticalElbowLink)
        self.ui_menuConvert.addAction(self.ui_actionToHorizontalElbowLink)
        self.ui_menuGraphics.addAction(self.ui_menuLock.menuAction())
        self.ui_menuGraphics.addAction(self.ui_menuConvert.menuAction())
        self.ui_menuGraphics.addSeparator()
        self.ui_menuGraphics.addAction(self.ui_actionSelectAll)
        self.ui_menuAlign.addAction(self.ui_actionAlignLeft)
        self.ui_menuAlign.addAction(self.ui_actionAlignCenter)
        self.ui_menuAlign.addAction(self.ui_actionAlignRight)
        self.ui_menuAlign.addSeparator()
        self.ui_menuAlign.addAction(self.ui_actionAlignTop)
        self.ui_menuAlign.addAction(self.ui_actionAlignMiddle)
        self.ui_menuAlign.addAction(self.ui_actionAlignBottom)
        self.ui_menuAlign.addSeparator()
        self.ui_menuAlign.addAction(self.ui_actionDistributeHorizontally)
        self.ui_menuAlign.addAction(self.ui_actionDistributeVertically)
        self.ui_menuLayout.addAction(self.ui_menuAlign.menuAction())
        self.ui_menubar.addAction(self.ui_menuFile.menuAction())
        self.ui_menubar.addAction(self.ui_menuEdit.menuAction())
        self.ui_menubar.addAction(self.ui_menuView.menuAction())
        self.ui_menubar.addAction(self.ui_menuGraphics.menuAction())
        self.ui_menubar.addAction(self.ui_menuLayout.menuAction())

        self.retranslateUi(MainWindow)
        self.ui_mainTabWidget.setCurrentIndex(-1)
        self.ui_consoleTabWidget.setCurrentIndex(-1)
        self.ui_styleTabWidget.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.ui_menuFile.setTitle(_translate("MainWindow", "File"))
        self.ui_menuView.setTitle(_translate("MainWindow", "View"))
        self.ui_menuEdit.setTitle(_translate("MainWindow", "Edit"))
        self.ui_menuGraphics.setTitle(_translate("MainWindow", "Graphics"))
        self.ui_menuLock.setTitle(_translate("MainWindow", "Lock"))
        self.ui_menuConvert.setTitle(_translate("MainWindow", "Convert"))
        self.ui_menuLayout.setTitle(_translate("MainWindow", "Layout"))
        self.ui_menuAlign.setTitle(_translate("MainWindow", "Align"))
        self.ui_navigationDock.setWindowTitle(_translate("MainWindow", "Navigation"))
        self.ui_groupNavigation.setTitle(_translate("MainWindow", "Networks"))
        self.ui_groupCode.setTitle(_translate("MainWindow", "From Python Code"))
        self.label.setText(_translate("MainWindow", "File"))
        self.ui_openCodeButton.setText(_translate("MainWindow", "Open"))
        self.ui_runButton.setText(_translate("MainWindow", "Run"))
        self.ui_consoleDock.setWindowTitle(_translate("MainWindow", "Console"))
        self.ui_visStyleDock.setWindowTitle(_translate("MainWindow", "Visual Style"))
        self.ui_styleTabWidget.setTabText(self.ui_styleTabWidget.indexOf(self.ui_nodeTabLayout), _translate("MainWindow", "Node"))
        self.ui_styleTabWidget.setTabText(self.ui_styleTabWidget.indexOf(self.ui_linkTabLayout), _translate("MainWindow", "Link"))
        self.ui_historyDock.setWindowTitle(_translate("MainWindow", "History"))
        self.ui_actionSaveNetwork.setText(_translate("MainWindow", "Save Network"))
        self.ui_actionExit.setText(_translate("MainWindow", "Exit"))
        self.ui_actionExit.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.ui_actionOpenNetwork.setText(_translate("MainWindow", "Open Network"))
        self.ui_actionNewView.setText(_translate("MainWindow", "New View"))
        self.ui_actionSaveSimulation.setText(_translate("MainWindow", "Save Simulation"))
        self.ui_actionOpenSimulation.setText(_translate("MainWindow", "Open Simulation"))
        self.ui_actionViewNavigationDock.setText(_translate("MainWindow", "Navigation"))
        self.ui_actionViewConsoleDock.setText(_translate("MainWindow", "Console"))
        self.ui_actionViewVisualStyleDock.setText(_translate("MainWindow", "Visual Style"))
        self.ui_actionExportImage.setText(_translate("MainWindow", "Export Image"))
        self.ui_actionCopy.setText(_translate("MainWindow", "Copy"))
        self.ui_actionPaste.setText(_translate("MainWindow", "Paste"))
        self.actionLock_Nodes.setText(_translate("MainWindow", "Lock Nodes"))
        self.actionLock_Links.setText(_translate("MainWindow", "Lock Links"))
        self.ui_actionLockNodes.setText(_translate("MainWindow", "Lock Nodes"))
        self.ui_actionLockLinks.setText(_translate("MainWindow", "Lock Links"))
        self.ui_actionLockLabels.setText(_translate("MainWindow", "Lock Labels"))
        self.ui_actionAlignLeft.setText(_translate("MainWindow", "Align Left"))
        self.ui_actionAlignCenter.setText(_translate("MainWindow", "Align Center"))
        self.ui_actionAlignRight.setText(_translate("MainWindow", "Align Right"))
        self.ui_actionAlignTop.setText(_translate("MainWindow", "Align Top"))
        self.ui_actionAlignMiddle.setText(_translate("MainWindow", "Align Middle"))
        self.ui_actionAlignBottom.setText(_translate("MainWindow", "Align Bottom"))
        self.ui_actionDistributeHorizontally.setText(_translate("MainWindow", "Distribute Horizontally"))
        self.ui_actionDistributeVertically.setText(_translate("MainWindow", "Distribute Vertically"))
        self.ui_actionSelectAll.setText(_translate("MainWindow", "Select All"))
        self.ui_actionToCurvedLink.setText(_translate("MainWindow", "To Curved Link"))
        self.ui_actionToStraightLink.setText(_translate("MainWindow", "To Straight Link"))
        self.ui_actionToVerticalElbowLink.setText(_translate("MainWindow", "To Vertical Elbow Link"))
        self.ui_actionToHorizontalElbowLink.setText(_translate("MainWindow", "To Horizontal Elbow Link"))
        self.ui_actionToEllipseNode.setText(_translate("MainWindow", "To Ellipse Node"))
        self.ui_actionToRectangleNode.setText(_translate("MainWindow", "To Rectangle Node"))
        self.ui_actionUndo.setText(_translate("MainWindow", "Undo"))
        self.ui_actionRedo.setText(_translate("MainWindow", "Redo"))
        self.ui_actionViewHistoryDock.setText(_translate("MainWindow", "History"))

from nezzle.widgets.dockwidget import DockWidget
from nezzle.widgets.historyview import HistoryView
from nezzle.widgets.networktreeview import NetworkTreeView



# [REF] http://stackoverflow.com/a/41070191

from qtpy.QtCore import Qt
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QMenu
from qtpy.QtWidgets import QTabWidget

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager


class ConsoleWidget(RichJupyterWidget):

    def __init__(self, *args,  **kwargs):
        super().__init__(*args, **kwargs)

        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel(show_banner=False)
        self.kernel_manager.kernel.gui = 'qt'
        self.kernel_manager.kernel.shell.cache_size = 50000

        def _abort_queues(kernel):
            pass

        self.kernel_manager.kernel._abort_queues = _abort_queues
        
        self.kernel_client = self.kernel_manager.client()
        self.kernel_client.start_channels()

        self.buffer_size = 1000
        self.exit_requested.connect(self.stop)

        self.setAttribute(Qt.WA_DeleteOnClose)


    @Slot()
    def stop(self):
        self.kernel_client.stop_channels()
        self.kernel_manager.shutdown_kernel()

        self.close()
        self.deleteLater()


    def pushVariables(self, variableDict):
        """
        Given a dictionary containing labels / value pairs, push those variables
        to the Jupyter console widget
        """
        self.kernel_manager.kernel.shell.push(variableDict)

    def clear(self):
        """
        Clears the terminal
        """
        
        self._control.clear()

    def printText(self, text):
        """
        Prints some plain text to the console
        """
        self._append_plain_text(text)

    def executeCommand(self, command):
        """
        Execute a command in the frame of the console widget
        """
        self._execute(command, False)
        
    def reset(self, clear=True):
        super().reset(clear)


class ConsoleTabWidget(QTabWidget):

        def __init__(self, *args, **kwargs):
        
            super().__init__(*args, **kwargs)
            self.setTabsClosable(True)
            self.setMovable(True)
            
            self.tabCloseRequested.connect(self.closeTab)
                        
            # Set button context menu policy
            self.setContextMenuPolicy(Qt.CustomContextMenu)
            self.customContextMenuRequested.connect(self.onContextMenu)

            # Create context menu
            self.pop_menu = QMenu(self)
            self.actionNewConsole = self.pop_menu.addAction('New console')
            self.appendConsole()
            
        def appendConsole(self):            
            console = ConsoleWidget()
            self.addTab(console, "Console")
            console.pushVariables({'console': console})

        @Slot()
        def onContextMenu(self, point):
            if self.count() == 1:
                self.actionNewConsole.setEnabled(False)
            elif self.count() == 0:
                self.actionNewConsole.setEnabled(True)
            else:
                raise RuntimeError("An illegal state of the ConsoleTabWidget object "
                                    "has been detected.")
            
            selectedAction = self.pop_menu.exec_(self.mapToGlobal(point))
            if selectedAction == self.actionNewConsole:
                self.appendConsole()                
                
            
        @Slot(int)
        def closeTab(self, currentIndex):   
            consoleWidget = self.widget(currentIndex)
            consoleWidget.stop()
            self.removeTab(currentIndex)



from qtpy.QtCore import Qt
from qtpy.QtGui import QColor

from nezzle.utils import Singleton

@Singleton
class AttributeMapper(object):
    def __init__(self):
        self._map_to_qt = {}

        # BORDER_JOIN
        self._map_to_qt['BORDER_JOIN'] = {}
        self._map_to_qt['BORDER_JOIN']['BEVEL'] = Qt.BevelJoin
        self._map_to_qt['BORDER_JOIN']['MITER'] = Qt.MiterJoin
        self._map_to_qt['BORDER_JOIN']['ROUND'] = Qt.RoundJoin

        # LINE_TYPE
        self._map_to_qt['BORDER_LINE'] = {}
        self._map_to_qt['BORDER_LINE']['SOLID'] = Qt.SolidLine
        self._map_to_qt['BORDER_LINE']['DASH'] = Qt.DashLine
        self._map_to_qt['BORDER_LINE']['DOT'] = Qt.DotLine
        self._map_to_qt['BORDER_LINE']['DASHDOT'] = Qt.DashDotLine
        self._map_to_qt['BORDER_LINE']['DASHDOTDOT'] = Qt.DashDotDotLine

        # Create a mapper vice versa.
        self._map_to_str = {}
        for attr, options in self._map_to_qt.items():
            self._map_to_str[attr] = {}
            d = {qtobj:strval for strval, qtobj in options.items()}
            self._map_to_str[attr] = d

    def to_qt(self, attr, val):
        """
        e.g.) to_qt('BORDER_JOIN', 'BEVEL') => Qt.BevelJoin
        e.g.) to_qt('BORDER_COLOR', '#000000') => QColor('#000000')

        """
        if not isinstance(val, str):
            raise ValueError("Converting to an Qt Object "
                             "requires Python string, not %s"%(val))

        attr = attr.upper()
        val = val.upper()
        if attr.endswith('_COLOR'):
            return self._to_qcolor(attr, val)

        if attr in self._map_to_qt:
            return self._map_to_qt[attr][val]
        else:
            raise AttributeError("It is not a valid attribute.")

    def to_str(self, attr, obj):

        attr = attr.upper()
        if attr.endswith('_COLOR'):
            return self._to_color_str(attr, obj)

        if attr in self._map_to_str:
            return self._map_to_str[attr][obj]
        else:
            raise AttributeError("It is not a valid attribute.")

    def _to_qcolor(self, attr, val):
        if not QColor.isValidColor(val):
            raise AttributeError("It is not a valid color string: %s"%(val))

        color = QColor(val)
        return color

    def _to_color_str(self, obj):
        if isinstance(obj, QColor):
            return obj.name(QColor.HexArgb)
        elif isinstance(obj, Qt.GlobalColor):
            color = QColor(obj)
            return color.name(QColor.HexArgb)
        else:
            raise AttributeError("It is not a valid Qt color object: %s"%(obj))




import numpy as np

from qtpy.QtCore import Qt
from qtpy.QtCore import Slot, Signal
from qtpy import QtWidgets
from qtpy.QtWidgets import QApplication
from qtpy.QtWidgets import QGraphicsScene
from qtpy.QtWidgets import QGraphicsView
from qtpy.QtWidgets import QRubberBand
from qtpy.QtWidgets import QMenu
from qtpy.QtGui import QKeyEvent
from qtpy.QtGui import QPixmapCache
from qtpy.QtGui import QPainter
from qtpy.QtGui import QTransform


from nezzle.history import History
from nezzle.graphics.nodes.basenode import BaseNode


class GraphicsView(QGraphicsView):

    def __init__(self, main_window=None, parent=None):
        super().__init__(parent)
        self.mw = main_window
        self.setDragMode(QGraphicsView.RubberBandDrag)
        self.setRenderHint(QPainter.Antialiasing)
        self.setRenderHint(QPainter.TextAntialiasing)
        self.rubberBand = QRubberBand(QRubberBand.Rectangle, self)

        # The default of cache mode is no cache (0)
        self.setCacheMode(QGraphicsView.CacheBackground)

        QPixmapCache.setCacheLimit(102400)
        self.scale(1.0, 1.0)

        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setViewportUpdateMode(QtWidgets.QGraphicsView.SmartViewportUpdate)

        self.setAcceptDrops(True)

        # Create context menu
        if self.mw:
            self.pop_menu = QMenu(self)
            self.pop_menu.addMenu(self.mw.ui_menuAlign)  # Use the existing one.
            self.pop_menu.addMenu(self.mw.ui_menuGraphics)

        self._pos_drag_start = None
        self._is_dragged = False
        self._item_clicked = None
        self._moved_by_key = False

    @Slot()
    def on_context_menu(self, event):
        self.pop_menu.exec_(self.mapToGlobal(event.pos()))

    def undo_current_scene(self):
        self.scene().history.undo()

    def redo_current_scene(self):
        self.scene().history.redo()

    def enable_menu_align(self):
        """Turn on/off the alignment functionality.
        """
        items = self.scene().selected_movable_items()

        if self.mw and len(items) > 1:
            self.mw.ui_menuAlign.setEnabled(True)
        else:
            self.mw.ui_menuAlign.setEnabled(False)

    def mousePressEvent(self, event):
        self._item_clicked = self.scene().itemAt(self.mapToScene(event.pos()), QTransform())
        if event.button() == Qt.LeftButton:
            self._pos_drag_start = event.pos()

            if self._item_clicked:
                super().mousePressEvent(event)
                items = self.scene().selectedItems()
                self.scene().update_selected_items_old_positions()
                return

        if self.dragMode() == QGraphicsView.RubberBandDrag:
            # Process the context menu actions such as alignment.
            if event.button() == Qt.RightButton:
                self.enable_menu_align()
                self.on_context_menu(event)
                # Should not return super().mousePressEvent(event) here.
                # Otherwise, the press event is processed under the drag mode.
                return

        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if event.modifiers() == Qt.ControlModifier:
            return super().mouseMoveEvent(event)

        if event.buttons() != Qt.LeftButton:
            return super().mouseMoveEvent(event)

        if (event.pos() - self._pos_drag_start).manhattanLength() < 0.1: #QApplication.startDragDistance():
            return super().mouseMoveEvent(event)

        self._is_dragged = True

        self.scene().update()
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self._item_clicked and self._is_dragged:
            items = self.scene().selected_movable_items()
            num_items = len(items)
            if num_items > 0:
                self.scene().history.on_move_items_by_mouse(items)

        # Updating the old positions should be done after emitting the signal to History object.
        if self.dragMode() == QGraphicsView.RubberBandDrag:
            self.mw.ct_manager.update_console_vars()
            self.enable_menu_align()
            self.scene().update_selected_items_old_positions()

        self._is_dragged = False
        self.scene().update()
        return super().mouseReleaseEvent(event)

    def wheelEvent(self, event):
        ad = -event.angleDelta().y()
        # factor = 1.41 ** (-ad / 240.0)
        if ad < 0:
            factor = 0.9
        else:
            factor = 1.1

        self.scale(factor, factor)
        return super().wheelEvent(event)

    def keyPressEvent(self, event):
        if type(event) == QKeyEvent:
            if event.key() == Qt.Key_Space:
                self.setInteractive(False)  # Not to select any item
                self.setDragMode(QGraphicsView.ScrollHandDrag)
                return

            # Move the selected items with arrow keys
            if event.key() in [Qt.Key_Up, Qt.Key_Down, Qt.Key_Left, Qt.Key_Right]:

                if not event.isAutoRepeat():
                    self.scene().update_selected_items_old_positions()

                items = self.scene().selected_movable_items()
                if len(items) <= 0:
                    return
                if event.key() == Qt.Key_Up:
                    for item in items:
                        item.setY(item.y() - 1)
                elif event.key() == Qt.Key_Down:
                    for item in items:
                        item.setY(item.y() + 1)
                elif event.key() == Qt.Key_Left:
                    for item in items:
                        item.setX(item.x() - 1)
                elif event.key() == Qt.Key_Right:
                    for item in items:
                        item.setX(item.x() + 1)

                self._moved_by_key = True
            # end of if

    def keyReleaseEvent(self, event):
        if type(event) == QKeyEvent:
            if event.key() == Qt.Key_Space:
                self.setDragMode(QGraphicsView.RubberBandDrag)
                self.setInteractive(True)

            if not event.isAutoRepeat() and event.key() in [Qt.Key_Up, Qt.Key_Down, Qt.Key_Left, Qt.Key_Right]:
                items = self.scene().selected_movable_items()
                num_items = len(items)
                if self._moved_by_key and num_items > 0:
                    self.scene().history.on_move_items_by_key(items)
                    self._moved_by_key = False

    def _list_pos_x(self, items):
        return [item.x() for item in items]

    def _list_pos_y(self, items):
        return [item.y() for item in items]

    def _list_pos(self, items):
        return [item.pos() for item in items]

    def _set_items_pos_x(self, items, aggfunc):
        pos_x = aggfunc(self._list_pos_x(items))
        for item in items:
            item.setX(pos_x)

    def _set_items_pos_y(self, items, aggfunc):
        pos_y = aggfunc(self._list_pos_y(items))
        for item in items:
            item.setY(pos_y)

    def align_objects(self, direction):
        items = self.scene().selected_movable_items()

        if direction == 'left':
            self._set_items_pos_x(items, min)
        elif direction == 'right':
            self._set_items_pos_x(items, max)
        elif direction == 'center':
            self._set_items_pos_x(items, np.average)
        elif direction == 'top':
            self._set_items_pos_y(items, min)
        elif direction == 'bottom':
            self._set_items_pos_y(items, max)
        elif direction == 'middle':
            self._set_items_pos_y(items, np.average)
        else:
            raise ValueError("Unknown direction for alignment: %s"%(direction))

        self.scene().history.on_move_items_by_mouse(items)

    def distribute_objects(self, direction):
        items = self.scene().selected_movable_items()

        if direction == 'horizontal':  # x
            items.sort(key=lambda obj: obj.x())
            x_max = max([item.x() for item in items])
            x_min = min([item.x() for item in items])
            gap_x = (x_max - x_min) / (len(items) - 1)
            for i, item in enumerate(items):
                item.setX(x_min + i*gap_x)
                item.update()

        elif direction == 'vertical':  # y
            items.sort(key=lambda obj: obj.y())
            y_max = max([item.y() for item in items])
            y_min = min([item.y() for item in items])
            gap_y = (y_max - y_min) / (len(items) - 1)
            for i, item in enumerate(items):
                item.setY(y_min + i*gap_y)
                item.update()

        else:
            raise ValueError("Unknown direction for distribution: %s"%(direction))

        self.scene().history.on_move_items_by_mouse(items)


class GraphicsScene(QGraphicsScene):

    def __init__(self, *args, parent=None, **kwargs):
        super().__init__(*args, parent, **kwargs)
        self.setBackgroundBrush(Qt.transparent)
        self.setItemIndexMethod(QGraphicsScene.NoIndex)

        self._history = History(self)

    @property
    def view(self):
        return self.views()[0]

    @property
    def history(self):
        return self._history

    def selected_movable_items(self):
        items_selected = self.selectedItems()
        return [item for item in items_selected if item.is_movable()]

    def update_selected_items_old_positions(self):
        items = self.selectedItems()  # Don't use selected_movable_items()
        if len(items) == 0:
            return

        for item in items:
            if item.is_movable():
                item["_OLD_POS"] = item.pos()

            item.update_children_old_positions()


# end of class



from qtpy.QtCore import Slot
from qtpy.QtCore import Qt
from qtpy.QtGui import QDoubleValidator
from qtpy.QtWidgets import QDialog, QFileDialog
from qtpy.QtWidgets import QMessageBox
from qtpy.QtWidgets import QTableWidgetItem
from qtpy.QtWidgets import QComboBox

from nezzle.ui.ui_opennetworkdialog import Ui_OpenNetworkDialog
from nezzle.fileio import read_metadata
from nezzle.graphics.arrows.arrowclassfactory import ArrowClassFactory


class OpenNetworkDialog(QDialog, Ui_OpenNetworkDialog):

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.setWindowTitle("Open a network file")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        validator_double = QDoubleValidator()
        validator_double.setBottom(0.0)

        # Set TableWidget for link haader mapping
        self.ui_linkHeadMappingTable.setColumnCount(3)
        self.ui_linkHeadMappingTable.setHorizontalHeaderLabels(["Count", "Interaction", "Head"])

        # Set head selection QComboBox
        self._available_heads = ArrowClassFactory.get_available_heads()
        self._head_selections = {}

        # Link signal and slot
        self.ui_openButton.released.connect(self.on_open_button_released)
        self.ui_reloadButton.released.connect(self.on_reload_button_released)
        self.ui_buttonBox.accepted.connect(self.on_accepted)
        self.ui_buttonBox.rejected.connect(self.on_rejected)

        # TODO: Implement tail mapping
        self.ui_linkTailMappingTable.setColumnCount(3)
        self.ui_linkTailMappingTable.setHorizontalHeaderLabels(["Count", "Interaction", "Tail"])
        self.ui_linkTailMappingTable.setItem(0, 0, QTableWidgetItem("Currently Not Supported"))

        self.ui_linkTailMappingTable.setDisabled(True)


    @Slot()
    def on_open_button_released(self):
        try:
            dialog = QFileDialog(self)
            dialog.setWindowTitle("Open a network file")
            dialog.setAcceptMode(QFileDialog.AcceptOpen)
            dialog.setNameFilters([self.tr("Text files (*.sif *.json)")])
            dialog.setFileMode(QFileDialog.ExistingFile)
            if dialog.exec() == QDialog.Accepted:
                fpath = dialog.selectedFiles()[0]
                self.ui_filePathEdit.setText(fpath)
                self.on_reload_button_released()
            # end of if
        except Exception as err:
            print(err)
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setWindowTitle("Open a network file")
            msg.setText("An invalid file type for network data.")
            msg.exec()
        # end of except

    @Slot()
    def on_reload_button_released(self):
        fpath = self.ui_filePathEdit.text()
        fpath = fpath.strip()

        if not fpath or len(fpath) <= 0:
            return

        self._head_selections.clear()
        metadata = read_metadata(fpath)
        interactions = metadata["INTERACTIONS"]
        num_rows = len(interactions)
        self.ui_linkHeadMappingTable.setRowCount(num_rows)
        for i, (interaction_name, count) in enumerate(interactions.items()):
            # Column (0): Count of each interaction
            count_item = QTableWidgetItem(str(count))
            count_item.setTextAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
            count_item.setFlags(count_item.flags() ^ Qt.ItemIsEditable)
            self.ui_linkHeadMappingTable.setItem(i, 0, count_item)

            # Column (1): Interaction name
            interaction_item = QTableWidgetItem(interaction_name)
            interaction_item.setTextAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
            count_item.setFlags(count_item.flags() ^ Qt.ItemIsEditable)
            self.ui_linkHeadMappingTable.setItem(i, 1, interaction_item)

            # Column (2): QComboBox for selecting heads
            cb = QComboBox()
            cb.addItems(self._available_heads)
            if interaction_name in self._available_heads:
                index = self._available_heads.index(interaction_name)
                cb.setCurrentIndex(index)

            self.ui_linkHeadMappingTable.setCellWidget(i, 2, cb)
            self._head_selections[interaction_name] = cb
        # end of for

        self.ui_networkNameEdit.setText(metadata["NETWORK_NAME"])

    @Slot()
    def on_accepted(self):
        self._fpath = self.ui_filePathEdit.text()
        self._network_name = self.ui_networkNameEdit.text().strip()
        super().accept()

    @Slot()
    def on_rejected(self):
        self.close()
        super().reject()

    # Read-only properties
    @property
    def link_map(self):
        _link_map = {}
        for interaction_name, cb in self._head_selections.items():
            _link_map[interaction_name] = cb.currentText()
        return _link_map

    @property
    def fpath(self):
        return self._fpath

    @property
    def network_name(self):
        return self._network_name

    # @property
    # def act_sym(self):
    #     return self._act_sym
    #
    # @property
    # def inh_sym(self):
    #     return self._inh_sym

    # @property
    # def scene_width(self):
    #     return self._scene_width
    #
    # @property
    # def scene_height(self):
    #     return self._scene_height

    @property
    def no_link_type(self):
        return self.ui_noLinkTypeCheck.isChecked()
# end of class